import copy
import os

import networkx as nx
import numpy as np
import pandas as pd
from IPython.display import SVG, Image, display

import trisicell as tsc
from trisicell.pl._annotation import _add_barplot, _add_chromplot, _get_tree
from trisicell.ul._trees import _info2_mutation_list, _to_newick


def clonal_tree(
    tree,
    muts_as_number=False,
    show_cells=True,
    collapsed_path=False,
    cell_info=None,
    output_file=None,
):
    """Draw the tree in clonal format.

    This functions plots the tree in which edges are mutations and nodes are cells.

    Parameters
    ----------
    tree : :class:`networkx.DiGraph`
        The input tree.
    muts_as_number : :obj:`bool`, optional
        Change the mutation list to a number at edges, by default False
    show_cells : :obj:`bool`, optional
        Whether to show cells or not, by default True
    collapsed_path : :obj:`bool`, optional
        Collapsed linear path in the tree , by default False
    cell_info : :class:`pandas.DataFrame`, optional
        Information of cells for coloring the nodes by a pie chart, by default None
    output_file : :obj:`str`, optional
        Path to a file for saving the tree in, by default None

    Returns
    -------
    :obj:`None`
    """

    tc = tree.copy()
    root = None
    for node in tc.nodes:
        if tc.in_degree(node) == 0:
            root = node
            break

    if collapsed_path:
        muts_as_number = True
        show_cells = True

    if muts_as_number:
        for u, v, l in tc.edges.data("label"):
            ll = l.split(tc.graph["splitter_mut"])
            if "––" in tc.nodes[v]["label"]:
                tc.add_edge(
                    u, v, label=f"  {len(ll)}  ", color="black", fontcolor="black"
                )
            else:
                tc.add_edge(u, v, label=f"  {len(ll)}  ")

    if not show_cells:
        for node in tc.nodes:
            if node != root:
                tc.nodes[node]["label"] = ""
                tc.nodes[node]["shape"] = "circle"
                tc.nodes[node]["width"] = 0
                tc.nodes[node]["height"] = 0

    if collapsed_path:
        tc2 = tc.copy()
        for _ in range(len(tc.nodes)):
            d_in = tc2.in_degree(tc2)
            d_out = tc2.out_degree(tc2)
            for node in tc2.nodes():
                if d_out[node] == 1 and d_in[node] == 1:
                    parent = [x for x in tc2.predecessors(node)][0]
                    child = [x for x in tc2.successors(node)][0]
                    if d_out[parent] < 2 and d_in[parent] == 1:
                        new_node = f"{parent}+{node}"
                        new_label = (
                            f"{tc2.nodes[parent]['label']}"
                            f"{tc.graph['splitter_cell']}"
                            f"{tc2.nodes[node]['label']}"
                        )
                        a = int(tc2[parent][node]["label"])
                        b = int(tc2[node][child]["label"])
                        new_edge = f"{a + b}"

                        tc2 = nx.contracted_nodes(tc2, parent, node, self_loops=False)
                        mapping = {parent: new_node}
                        tc2 = nx.relabel_nodes(tc2, mapping)
                        tc2[new_node][child]["label"] = new_edge
                        tc2.nodes[new_node]["label"] = new_label
                        break
        d_in = tc2.in_degree(tc2)
        d_out = tc2.out_degree(tc2)
        nodes = []
        for node in tc2.nodes():
            if d_out[node] == 0:
                nodes.append(node)
        for node in nodes:
            parent = [x for x in tc2.predecessors(node)][0]
            if d_out[parent] == 1 and d_in[parent] == 1:
                grandparent = [x for x in tc2.predecessors(parent)][0]

                new_node = f"{parent}+{node}"
                new_label = (
                    f"{tc2.nodes[parent]['label']}"
                    f"{tc.graph['splitter_cell']}"
                    f"{tc2.nodes[node]['label']}"
                )
                a = int(tc2[grandparent][parent]["label"])
                b = int(tc2[parent][node]["label"])
                new_edge = f"{a + b}"

                tc2 = nx.contracted_nodes(tc2, parent, node, self_loops=False)
                mapping = {parent: new_node}
                tc2 = nx.relabel_nodes(tc2, mapping)
                tc2[grandparent][new_node]["label"] = new_edge
                tc2.nodes[new_node]["label"] = new_label
        tc = tc2

        mapping = cell_info["name"].to_dict()
        for node in tc:
            if node != root:
                num = 0
                paths = nx.shortest_path(tc, source=root, target=node)
                for i in range(len(paths) - 1):
                    x = paths[i]
                    y = paths[i + 1]
                    num += len(tc[x][y]["label"].split(tc.graph["splitter_mut"]))
                try:
                    freq = [
                        mapping[x].split("_")[1]
                        for x in tc.nodes[node]["label"].split(
                            tc.graph["splitter_cell"]
                        )
                    ]
                except:
                    freq = ["#FFFFFF"]
                freq = pd.DataFrame(freq)[0].value_counts(normalize=True)
                fillcolor = ""
                for index, value in freq.items():
                    fillcolor += f"{index};{value}:"
                tc.nodes[node]["fontsize"] = 0.5
                tc.nodes[node]["shape"] = "circle"
                tc.nodes[node]["fontname"] = "Helvetica"
                tc.nodes[node]["style"] = "wedged"
                tc.nodes[node]["margin"] = 0.05
                tc.nodes[node]["pad"] = 0
                if "––" in tc.nodes[node]["label"]:
                    tc.nodes[node]["width"] = 0
                    tc.nodes[node]["height"] = 0
                else:
                    tc.nodes[node]["width"] = 0.8
                    tc.nodes[node]["height"] = 0.8
                tc.nodes[node]["fillcolor"] = fillcolor  # "red;0.3:green;0.6:orange"
                tc.nodes[node]["label"] = node
            else:
                tc.nodes[node]["label"] = f"<<b>zygote</b>>"

    if output_file is not None:
        mygraph = nx.drawing.nx_agraph.to_agraph(tc)
        mygraph.layout(prog="dot")
        mygraph.draw(output_file)

    return display(SVG(nx.drawing.nx_pydot.to_pydot(tc).create_svg()))


def _merge(x):
    return (
        f"{x['Ensemble']}_{x['Gene']}.{x['Chrom']}."
        f"{x['Position']}.{x['Reference']}.{x['Alteration']}"
    )


def _get_newick_info2_mutations(tree):
    newick = _to_newick(tree)
    info2, mutation_list = _info2_mutation_list(tree)
    return newick, info2, mutation_list


def dendro_tree(
    tree,
    width=8,
    height=3,
    cell_info=None,
    label_color="group_color",
    line_size=0.4,
    tiplab_size=2.5,
    inner_node_id=False,
    inner_node_size=2,
    distance_labels_to_bottom=4,
    annotation=[],
):
    """Draw the tree in dendro fromat.

    Parameters
    ----------
    tree : :class:`networkx.DiGraph`
        The input tree.
    width : :obj:`int`, optional
        Width of the figure, by default 8
    height : :obj:`int`, optional
        Height of the figure, by default 3
    cell_info : :class:`pandas.DataFrame`, optional
        Information about cells such as color,
        expression values of genes and etc, by default None
    label_color : :obj:`str`, optional
        The column name in which colors of cells are stored
        in the dataframe provided as `cell_info`, by default "group_color"
    line_size : :obj:`float`, optional
        Line size of the tree, by default 0.4
    tiplab_size : :obj:`float`, optional
        Cell name size in the tree, by default 2.5
    inner_node_id : :obj:`bool`, optional
        Information of the inner nodes (i.e. mutations), by default False
        Values are:

            - `nmuts_label`: only number of mutations
            - `nodeid_label`: only node id
            - `mixed_label`: both number of mutations and node id
    inner_node_size : :obj:`int`, optional
        Size of the inner nodes (i.e. mutations), by default 2
    distance_labels_to_bottom : :obj:`int`, optional
        Distance cell names to the bottom of the figure, by default 4
    annotation : :obj:`list`, optional
        List of gene names provided in the column dataframe of ``cell_info``
        in to be annotated in the bottom of the tree, by default []

    Returns
    -------
    :obj:`None`

    Raises
    ------
    SystemExit
        If `rpy2 <https://pypi.org/project/rpy2/>`_ is not already installed.
    SystemExit
        If `ggtree <https://github.com/YuLab-SMU/ggtree>`_ is not already installed.
    """

    try:
        import rpy2.robjects as robjects
        from rpy2.robjects import pandas2ri
        from rpy2.robjects.lib import grdevices
        from rpy2.robjects.packages import importr
    except:
        raise SystemExit(
            "A problem was encountered importing `rpy2`. "
            "To run this `rpy2` and `R` need to be installed."
        )
    try:
        importr("ggtree")
    except:
        raise SystemExit(
            "A problem was encountered importing `ggtree` in R. "
            "To run this `ggtree` needs to be installed in R. "
            "Use the following lines to installed them.\n\n"
            "devtools::install_github(c('YuLab-SMU/ggtree',"
            "'xiangpin/ggtreeExtra','YuLab-SMU/aplot'))\n"
            "install.packages('cowplot')\n"
        )

    newick, info2, mutation_list = _get_newick_info2_mutations(tree)
    mutation_list["index"] = mutation_list.apply(_merge, axis=1)
    tree.graph["mutation_list"] = mutation_list.set_index("index")
    tree.graph["newick"] = newick

    ggplot2 = importr("ggplot2")
    cowplot = importr("cowplot")
    ggtree = importr("ggtree")
    ape = importr("ape")
    aplot = importr("aplot")
    pandas2ri.activate()

    if cell_info is not None:
        if "group_color" not in cell_info.columns:
            cell_info["group_color"] = tree.graph["data"].shape[0] * ["#000000"]
        cell_info_r = robjects.conversion.py2rpy(cell_info.reset_index())
    else:
        cell_info_r = robjects.conversion.py2rpy(
            pd.DataFrame.from_dict(
                {
                    "x": tree.graph["data"].index,
                    "group_color": tree.graph["data"].shape[0] * ["#000000"],
                }
            )
        )
    robjects.globalenv["info1"] = cell_info_r
    info2_r = robjects.conversion.py2rpy(info2)
    robjects.globalenv["info2"] = info2_r

    # p = ggtree.ggtree(
    #     ape.read_tree(text=f"{newick}"), layout="dendrogram", size=line_size
    # )
    cmd = _get_tree(
        newick,
        line_size,
        label_color,
        tiplab_size,
        inner_node_id,
        inner_node_size,
        distance_labels_to_bottom,
    )

    for ann in annotation:
        cmd += "info1$id_index <- p$data[match(info1[,1], p$data$label),]$y"
        if ann[0] == "bar":
            cmd += _add_barplot(ann[1], ann[2])
        elif ann == "chrom":
            cmd += _add_chromplot()
        elif ann == "imputation":
            for imp in [
                ("f_0_1", "f_0_1_color"),
                ("f_1_0", "f_1_0_color"),
                ("f_3_0", "f_3_0_color"),
                ("f_3_1", "f_3_1_color"),
            ]:
                cmd += _add_barplot(imp[0], imp[1])

    with grdevices.render_to_bytesio(
        grdevices.svg, width=width, height=height
    ) as image:
        p = robjects.r(cmd)
        robjects.r.show(p)
        # robjects.r.ggsave(
        #     plot=p,
        #     filename="x.pdf",
        #     device="pdf",
        #     width=7.5,
        #     height=3,
        #     units="in",
        #     limitsize=False,
        # )
    return display(SVG(image.getvalue()))



import os
import time

import apted
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import trisicell as tsc
from trisicell.ul._trees import _split_labels, _to_apted, _to_mutation_tree


def ad(df_grnd, df_sol):
    """Ancestor-descendent accuracy.

    For each pair of mutations in ground truth tree that are in
    ancestor-descendant relation (same nodes excluded) we check
    whether this relationship is preserved in the inferred tree.

    This measure was introduced in :cite:`B-SCITE`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        The score is out of 1.00
    """

    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    M_grnd = df_grnd[inter].values
    M_sol = df_sol[inter].values
    error_pairs = []
    n_adpairs = 0
    for i in range(M_grnd.shape[1]):
        for j in range(i, M_grnd.shape[1]):
            cap1 = M_grnd[:, i] * M_grnd[:, j]
            cap2 = M_sol[:, i] * M_sol[:, j]
            if np.sum(cap1) > 0 and np.sum(M_grnd[:, i]) != np.sum(M_grnd[:, j]):
                n_adpairs = n_adpairs + 1
                if np.sum(cap2) == 0:
                    error_pairs.append([i, j])
                else:
                    if np.sum(M_grnd[:, j]) > np.sum(M_grnd[:, i]) and np.sum(
                        M_sol[:, j]
                    ) <= np.sum(M_sol[:, i]):
                        error_pairs.append([i, j])
                    else:
                        if np.sum(M_grnd[:, i]) > np.sum(M_grnd[:, j]) and np.sum(
                            M_sol[:, i]
                        ) <= np.sum(M_sol[:, j]):
                            error_pairs.append([i, j])
    return 1 - len(error_pairs) / n_adpairs


def dl(df_grnd, df_sol):
    """Different-lineage accuracy.

    For each pair of mutations in ground truth tree that are
    in different-lineages relation we check whether the same relationship
    is preserved in the inferred tree.

    This measure was introduced in :cite:`B-SCITE`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        The score is out of 1.00
    """

    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    M_grnd = df_grnd[inter].values
    M_sol = df_sol[inter].values
    error_pairs = []
    n_dlpairs1 = 0
    n_dlpairs2 = 0
    for i in range(M_grnd.shape[1]):
        for j in range(i, M_grnd.shape[1]):
            cap1 = M_grnd[:, i] * M_grnd[:, j]
            cap2 = M_sol[:, i] * M_sol[:, j]
            if (
                np.sum(cap1) == 0
                and np.sum(M_grnd[:, i]) != 0
                and np.sum(M_grnd[:, j]) != 0
            ):
                n_dlpairs1 = n_dlpairs1 + 1
                if (
                    np.sum(cap2) == 0
                    and np.sum(M_sol[:, i]) != 0
                    and np.sum(M_sol[:, j]) != 0
                ):
                    n_dlpairs2 = n_dlpairs2 + 1
    return n_dlpairs2 / n_dlpairs1


def cc(df_grnd, df_sol):
    """Co-clustering accuracy.

    For each pair of mutations in ground truth tree that are on the same node we look
    relationship is preserved in the inferred tree.

    This measure was introduced in :cite:`B-SCITE`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`float`
        The score is out of 1.00
    """

    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    M_grnd = df_grnd[inter].values
    M_sol = df_sol[inter].values

    # TODO: implement.
    pass


def mltd(df_grnd, df_sol):
    """Multi-labeled tree dissimilarity measure (MLTD).

    This measure was introduced in :cite:`MLTD`.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`int`
        MLTD
    :obj:`int`
        MLTS
    :obj:`float`
        MLTSM out of 1.00
    """

    def _convert_tree_to_mtld_input(tree, file):
        with open(file, "w") as fout:
            for u, v, l in tree.edges.data("label"):
                if tree.in_degree(u) == 0:
                    fout.write(f"{u}=\n")
                muts = l.split(tree.graph["splitter_mut"])
                fout.write(f"{v}={','.join(muts)}\n")
            for u in tree.nodes:
                children = [str(n) for n in tree.neighbors(u)]
                if len(children) == 0:
                    continue
                fout.write(f"{u}:{','.join(children)}\n")

    tmpdir = tsc.ul.tmpdir(prefix="trisicell.", suffix=".mltd", dirname=".")

    df_grnd.columns = df_grnd.columns.str.replace(":", "_").str.replace("=", "_")
    df_sol.columns = df_sol.columns.str.replace(":", "_").str.replace("=", "_")
    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = tsc.ul.to_tree(df_grnd1)
    tree_sol = tsc.ul.to_tree(df_sol1)

    _convert_tree_to_mtld_input(tree_grnd, f"{tmpdir}/grnd.in")
    _convert_tree_to_mtld_input(tree_sol, f"{tmpdir}/sol.in")

    mltd = tsc.ul.get_file("trisicell.external/bin/mltd")
    cmd = f"{mltd} {tmpdir}/grnd.in {tmpdir}/sol.in > {tmpdir}/res.out 2>&1"
    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    distance, similarity, mltsm = None, None, None
    with open(f"{tmpdir}/res.out") as fin:
        for line in fin:
            if line.startswith("Distance = "):
                distance = int(line.strip().replace("Distance = ", ""))
            if line.startswith("Similarity = "):
                similarity = int(line.strip().replace("Similarity = ", ""))
            if line.startswith("Normalized Similarity = "):
                mltsm = float(line.strip().replace("Normalized Similarity = ", ""))

    tsc.ul.cleanup(tmpdir)

    return distance, similarity, mltsm


def tpted(df_grnd, df_sol):
    """Tumor phylogeny tree edit distance measure (TPTED).

    This measure was introduced in :cite:`PhISCS`. This implementation uses
    `APTED <https://github.com/JoaoFelipe/apted>`_.

    Parameters
    ----------
    df_grnd : :class:`pandas.DataFrame`
        The first genotype matrix (e.g. ground truth)
        This matrix must be conflict-free.
    df_sol : :class:`pandas.DataFrame`
        The second genotype matrix (e.g. solution/inferred)
        This matrix must be conflict-free.

    Returns
    -------
    :obj:`int`
        distance
    """

    inter = np.intersect1d(df_grnd.columns, df_sol.columns)
    df_grnd1 = df_grnd[inter]
    df_sol1 = df_sol[inter]

    tree_grnd = tsc.ul.to_tree(df_grnd1)
    tree_sol = tsc.ul.to_tree(df_sol1)

    mt_grnd = _to_mutation_tree(tree_grnd)
    mt_sol = _to_mutation_tree(tree_sol)

    sl_grnd = _split_labels(mt_grnd, mt_guide=mt_sol)
    sl_sol = _split_labels(mt_sol, mt_guide=mt_sol)

    apted_grnd = _to_apted(sl_grnd)
    apted_sol = _to_apted(sl_sol)

    tree1 = apted.helpers.Tree.from_text(apted_grnd)
    tree2 = apted.helpers.Tree.from_text(apted_sol)

    ap = apted.APTED(tree1, tree2)
    ed = ap.compute_edit_distance()

    return 1 - ed / (2 * len(inter))



import pytest
from _helpers import *

import trisicell as tsc


class TestConsensus:
    def test_consensus_1(self):
        # result in biorxiv.figs18b
        f1 = tsc.ul.get_file("trisicell.datasets/test/consensus/biorxiv.fig3b.CFMatrix")
        f2 = tsc.ul.get_file(
            "trisicell.datasets/test/consensus/biorxiv.figs18a.CFMatrix"
        )
        sc1 = tsc.io.read(f1)
        sc2 = tsc.io.read(f2)
        final_tree = tsc.tl.consensus_tree(sc1, sc2)
        assert len(final_tree.nodes) == 19

    def test_consensus_2(self):
        # result in recomb.fig1c
        f1 = tsc.ul.get_file("trisicell.datasets/test/consensus/recomb.fig1a.CFMatrix")
        f2 = tsc.ul.get_file("trisicell.datasets/test/consensus/recomb.fig1b.CFMatrix")
        sc1 = tsc.io.read(f1)
        sc2 = tsc.io.read(f2)
        final_tree = tsc.tl.consensus_tree(sc1, sc2)
        assert len(final_tree.nodes) == 21

    def test_consensus_3(self):
        # result in biorxiv.fig4d
        f1 = tsc.ul.get_file("trisicell.datasets/test/consensus/biorxiv.fig4b.CFMatrix")
        f2 = tsc.ul.get_file("trisicell.datasets/test/consensus/biorxiv.fig4c.CFMatrix")
        sc1 = tsc.io.read(f1)
        sc2 = tsc.io.read(f2)
        final_tree = tsc.tl.consensus_tree(sc1, sc2)
        assert len(final_tree.nodes) == 34

    def test_consensus_4(self):
        # result in biorxiv.fig3f
        f1 = tsc.ul.get_file("trisicell.datasets/test/consensus/biorxiv.fig3b.CFMatrix")
        f2 = tsc.ul.get_file("trisicell.datasets/test/consensus/biorxiv.fig3c.CFMatrix")
        sc1 = tsc.io.read(f1)
        sc2 = tsc.io.read(f2)
        final_tree = tsc.tl.consensus_tree(sc1, sc2)
        assert len(final_tree.nodes) == 19



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: May 16, 2020
# Description: creating a new reference index based on the first pass run of the STAR
# =========================================================================================

from trisicell.ul._servers import *


def run03(config, afterok):
    def cmds():
        cmds = ""
        cmds += cmd([f"mkdir -p {config['outdir']}/_indexing/2"])
        cmds += cmd([f"module load STAR/2.7.3a"])
        files = " ".join(
            [f"{config['outdir']}/{s}/SJ.out.tab" for s in config["samples"]]
        )
        cmds += cmd(
            [
                f"STAR",
                f"--runMode genomeGenerate",
                f"--genomeDir {config['outdir']}/_indexing/2",
                f"--genomeFastaFiles {config['ref']}",
                f"--sjdbGTFfile {config['annot']}",
                f"--sjdbFileChrStartEnd {files}",
                f"--sjdbOverhang {config['readlength']}",
                f"--runThreadN 32",
            ]
        )
        cmds += cmd([f"echo Done!"], islast=True)
        return cmds

    df = pd.DataFrame()
    df["cmd"] = [cmds()]

    jobname = os.path.basename(__file__)[:-3]
    cmdmain = write_cmds_get_main(
        df,
        jobname,
        config["time03"],
        config["mem03"],
        "",
        32,
        config["email"],
        config["tmpdir"],
        afterok,
    )
    return cmdmain



import trisicell.datasets as datasets
import trisicell.io as io
import trisicell.logging as logg
import trisicell.pl as pl
import trisicell.pp as pp
import trisicell.settings as settings
import trisicell.tl as tl
import trisicell.ul as ul

__author__ = ", ".join(["Farid Rashidi"])
__maintainer__ = ", ".join(["Farid Rashidi"])
__email__ = ", ".join(["farid.rsh@gmail.com"])
__version__ = "0.0.3"



import os

import click

import trisicell as tsc


@click.command(short_help="Run HUNTRESS.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--method",
    "-m",
    default="dna",
    type=click.Choice(["dna", "fn", "rna"]),
    show_default=True,
    help="Method of the HUNTRESS",
)
@click.option(
    "--n_threads",
    "-p",
    default=1,
    type=int,
    show_default=True,
    help="Number of threads.",
)
def huntress(genotype_file, alpha, beta, method, n_threads):
    """HUNTRESS

    trisicell huntress input.SC 0.0001 0.1 -m dna -p 8
    """

    outfile = os.path.splitext(genotype_file)[0]

    tsc.settings.verbosity = "info"
    tsc.settings.logfile = f"{outfile}.huntress.log"

    df_out = tsc.tl.huntress(
        genotype_file, alpha=alpha, beta=beta, kind=method, n_threads=n_threads
    )
    tsc.io.write(df_out, f"{outfile}.huntress.CFMatrix")

    return None



import pandas as pd

import trisicell as tsc


def huntress(df_input_filepath, alpha, beta, kind, n_threads=1, save_inter=False):
    tsc.logg.info(
        f"running HUNTRESS with alpha={alpha}, beta={beta}, kind={kind}, n_threads={n_threads}"
    )
    tmpdir = tsc.ul.tmpdir(prefix="trisicell.", suffix=".huntress", dirname=".")

    # TODO remove tmpdir and directly work with python api.
    from ._huntress_21_03_19 import Reconstruct

    running_time = 0
    if kind == "dna":
        fn_conorm = 0.1
        fp_conorm = fn_conorm * alpha / beta
        fnfp_conorm = fn_conorm / fp_conorm

        running_time = Reconstruct(
            df_input_filepath,
            f"{tmpdir}/huntress",
            n_proc=n_threads,
            fnfp=51,
            post_fn=fn_conorm,
            post_fp=fp_conorm,
        )
        df_output = pd.read_table(f"{tmpdir}/huntress_optH_.CFMatrix", index_col=0)
    elif kind == "fn":
        running_time = Reconstruct(
            df_input_filepath,
            f"{tmpdir}/huntress.CFMatrix",
            Algchoice="FN",
            n_proc=n_threads,
        )
        df_output = pd.read_table(f"{tmpdir}/huntress.CFMatrix", index_col=0)
    elif kind == "rna":
        running_time = Reconstruct(
            df_input_filepath, f"{tmpdir}/huntress", postprocessing=1, n_proc=n_threads
        )
        df_output = pd.read_table(
            f"{tmpdir}/huntress_optH_.processed.CFMatrix", index_col=0
        )

    df_input = pd.read_table(df_input_filepath, index_col=0)
    tsc.ul.stat(df_input, df_output, alpha, beta, running_time)

    if not save_inter:
        tsc.ul.cleanup(tmpdir)
    return df_output



import math
import os
import random

import numpy as np
import pandas as pd

import trisicell as tsc


def simulate(n_cells=10, n_muts=10, n_clones=3, alpha=0.00001, beta=0.1, missing=0):
    """Simulate single-cell noisy genotype matrix.

    This function is using :cite:`OncoNEM`.

    Parameters
    ----------
    n_cells : :obj:`int`, optional
        Number of cells, by default 10
    n_muts : :obj:`int`, optional
        Number of mutations, by default 10
    n_clones : :obj:`int`, optional
        Number of clones, by default 3
    alpha : :obj:`float`, optional
        False positive rate, by default 0.00001
    beta : :obj:`float`, optional
        False negative rate, by default 0.1
    missing : :obj:`int`, optional
        Missing entry rate, by default 0

    Returns
    -------
    :class:`pandas.DataFrame`
        A genotype matrix where 0 is absent, 1 is present and 3 is missing.
    """

    onconem, onconem_is_not_imported = tsc.ul.import_rpy2(
        "oncoNEM",
        "BiocManager::install('graph')\ndevtools::install_bitbucket('edith_ross/oncoNEM')\n",
    )
    if onconem_is_not_imported:
        raise RuntimeError("Unable to import a package!")

    import rpy2.robjects as ro
    from rpy2.robjects import pandas2ri

    dat = onconem.simulateData(
        N_cells=n_cells,
        N_normalContam=0,
        N_clones=n_clones,
        N_unobs=0,
        N_sites=n_muts,
        FPR=alpha,
        FNR=beta,
        p_missing=missing,
        randomizeOrder=False,
    )

    with ro.conversion.localconverter(ro.default_converter + pandas2ri.converter):
        dat = ro.conversion.rpy2py(dat.rx2("D"))
    dat[dat == 2] = 3
    df = pd.DataFrame(dat, dtype=int)
    df.columns = [f"mut{x}" for x in df.columns]
    df.index = [f"cell{x}" for x in df.index]

    return df


def splatter(noisy, ground, seed=5):
    params = tsc.ul.get_file("trisicell.datasets/data/splatter.rds")

    splatter, splatter_is_not_imported = tsc.ul.import_rpy2(
        "splatter",
        "BiocManager::install('splatter')\n",
    )
    if splatter_is_not_imported:
        raise RuntimeError("Unable to import a package!")

    import rpy2.robjects as ro

    cmd = f"""
    suppressPackageStartupMessages({{
        library(splatter)
    }})

    seed <- strtoi({seed})
    name <- basename('{noisy}')
    groundmatrix <- read.table(file='{ground}', header=TRUE, sep='\t', row.names=1)
    counts <- colSums(groundmatrix == 1)
    # print(length(counts[counts < 2]))

    inmatrix <- read.table(file='{noisy}', header=TRUE, sep='\t', row.names=1)
    # counts <- colSums(inmatrix == 1)
    # print(counts[counts < 2])

    n <- dim(inmatrix)[1]
    m <- dim(inmatrix)[2]

    params <- readRDS(file='{params}')
    sim <- splatSimulate(params, nGenes=m, batchCells=n, seed=seed, verbose=FALSE)
    
    outmatrix <- inmatrix
    outmatrix[t(counts(sim)) < 10] <- 3

    counts <- colSums(outmatrix == 1)
    counts <- 2-counts[counts < 2]
    names <- names(counts)
    i<-0
    for (n in names) {{
        i <- i+1
        x <- which(inmatrix[,names[i]] == 1)
        if (counts[i] <= length(x)) {{
            my_sample <- sample(x, length(x), replace=FALSE)
            j <- 0
            for (s in my_sample) {{
                if (outmatrix[s,names[i]] == 3) {{
                    j <- j+1
                    outmatrix[s,names[i]] <- 1
                    if (counts[i] == j) {{
                        break
                    }}
                }}
            }}
        }}
    }}

    counts <- colSums(outmatrix == 1)
    counts <- 2-counts[counts < 2]
    names <- names(counts)
    i<-0
    for (n in names) {{
        i <- i+1
        x <- which(groundmatrix[,names[i]] == 1)
        # x2 <- which(inmatrix[,names[i]] == 0)
        # x <- intersect(x1, x2)
        if (counts[i] <= length(x)) {{
            my_sample <- sample(x, length(x), replace=FALSE)
            j <- 0
            for (s in my_sample) {{
                if (outmatrix[s,names[i]] == 0 || outmatrix[s,names[i]] == 3) {{
                    j <- j+1
                    outmatrix[s,names[i]] <- 1
                    if (counts[i] == j) {{
                        break
                    }}
                }}
            }}
        }}
    }}

    counts <- colSums(outmatrix == 1)
    # print(length(counts[counts < 2]))


    outfile <- '{noisy[:-len('.SC')]}-seed_{seed}.SC'
    write.table(x=data.frame('cellID_mutID'=rownames(outmatrix),outmatrix),
                             file=outfile, sep='\\t', quote=FALSE, row.names=FALSE)
    """
    ro.r(cmd)


def create_splatter():
    params = tsc.ul.get_file("trisicell.datasets/data/splatter.rds")

    splatter, splatter_is_not_imported = tsc.ul.import_rpy2(
        "splatter",
        "BiocManager::install('splatter')\n",
    )
    if splatter_is_not_imported:
        raise RuntimeError("Unable to import a package!")

    import rpy2.robjects as ro

    cmd = f"""
    suppressPackageStartupMessages({{
        library(splatter)
    }})

    readcountfile <- '~/122/expr.count.tsv'
    data <- read.table(file=readcountfile, header=TRUE, sep='\t', row.names=1)
    data <- t(as.matrix(data))
    params <- splatEstimate(data)
    saveRDS(params, file='{params}')
    """
    ro.r(cmd)


def add_noise(df_in, alpha, beta, missing):
    """Add noise to the input genotype matrix.

    These noise includes:
    1) False positive errors (alpha)
    2) False negative errors (beta)
    3) Missing entry errors (missing)

    Parameters
    ----------
    df_in : :class:`pandas.DataFrame`
        Input genotype matrix.
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    missing : :obj:`float`
        Missing entry error rate.

    Returns
    -------
    :class:`pandas.DataFrame`
        A noisy genotype matrix where 0 is absent, 1 is present and 3 is missing.
    """

    def toss(p):
        return True if random.random() < p else False

    data = df_in.values
    n, m = df_in.shape
    data2 = -1 * np.ones(shape=(n, m)).astype(int)
    countFP = 0
    countFN = 0
    countNA = 0
    countOneZero = 0
    indexNA = []
    changedBefore = []
    for i in range(n):
        for j in range(m):
            indexNA.append([i, j])
            countOneZero = countOneZero + 1
    random.shuffle(indexNA)
    nas = math.ceil(countOneZero * missing)
    for i in range(int(nas)):
        [a, b] = indexNA[i]
        changedBefore.append([a, b])
        data2[a][b] = 3
        countNA = countNA + 1
    for i in range(n):
        for j in range(m):
            if data2[i][j] != 3:
                if data[i][j] == 1:
                    if toss(beta):
                        data2[i][j] = 0
                        countFN = countFN + 1
                    else:
                        data2[i][j] = data[i][j]
                elif data[i][j] == 0:
                    if toss(alpha):
                        data2[i][j] = 1
                        countFP = countFP + 1
                    else:
                        data2[i][j] = data[i][j]
                else:
                    tsc.logg.error("Wrong Input")
                    sys.exit(2)

    tsc.logg.info(f"FNs={countFN}, FPs={countFP}, NAs={countNA}")

    df_out = pd.DataFrame(data2)
    df_out.columns = df_in.columns
    df_out.index = df_in.index
    df_out.index.name = "cellIDxmutID"

    return df_out



import os

import trisicell as tsc


def dendro(adata, kind):
    """Running DENDRO

    https://doi.org/10.1186/s13059-019-1922-x

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        Input data contains layers of mutant and total.
    kind : :obj:`str`
        Which dataset is this, either `brna` or `scrna`
    """

    try:
        import rpy2.robjects as robjects
        from rpy2.robjects.packages import importr
    except:
        raise SystemExit(
            "A problem was encountered importing `rpy2`. "
            "To run this `rpy2` and `R` need to be installed."
        )
    try:
        importr("DENDRO")
    except:
        raise SystemExit(
            "A problem was encountered importing `DENDRO` in R. "
            "To run this `DENDRO` needs to be installed in R. "
            "Use the following lines to installed them.\n\n"
            "devtools::install_github('zhouzilu/DENDRO')\n"
        )

    tsc.logg.info(f"running DENDRO")
    tmpdir = tsc.ul.tmpdir(prefix="trisicell.", suffix=".dendro", dirname=".")

    Z = adata.to_df(layer="genotype")
    Z[(Z == 1) | (Z == 3)] = 1
    Z[Z == 2] = 0
    X = adata.to_df(layer="mutant")
    N = adata.to_df(layer="total")
    Z.to_csv(os.path.join(tmpdir, "Z.csv"))
    X.to_csv(os.path.join(tmpdir, "X.csv"))
    N.to_csv(os.path.join(tmpdir, "N.csv"))

    if kind == "brna":
        width = 3.5
        height = 2
        body1 = f"""
        file <- './trisicell_notebooks/data/mouse.melanoma/24.bulk.rna.lcbg.csv'
        colors <- c('#D92347','#FF7F00','#812C7C','#009949','#0067A5')
        """

        body2 = f"""
        p <- ggtree(tree, layout='dendrogram', size=0.4) %<+% info1
        p <- p + geom_tiplab(aes(color=group), align=TRUE,
            linesize=.1, angle=-90, hjust=-0.1, size=2.5, fontface='bold')
        p <- p + scale_color_manual(values=colors, guide='none')
        p <- p + xlim_tree(20)
        """
    elif kind == "scrna":
        width = 8
        height = 3
        body1 = f"""
        file <- './trisicell_notebooks/data/mouse.melanoma/24.sc.rna.lcbg.csv'
        """
        body2 = f"""
        p <- ggtree(tree, layout='dendrogram', size=0.4) %<+% info1
        p <- p + geom_tiplab(aes(color=subclone_color), align=TRUE,
            linesize=.1, angle=-90, hjust=-0.1, size=1.5, fontface='bold')
        p <- p + scale_color_identity(guide='none')
        p <- p + xlim_tree(5)
        """
    else:
        width = 5
        height = 2
        body1 = f"""
        """
        body2 = f"""
        """

    cmd = f"""
    suppressMessages(library(DENDRO))
    suppressMessages(library(data.table))

    N <- t(as.matrix(fread('{tmpdir}/N.csv'), rownames=1))
    X <- t(as.matrix(fread('{tmpdir}/X.csv'), rownames=1))
    Z <- t(as.matrix(fread('{tmpdir}/Z.csv'), rownames=1))

    dist = DENDRO.dist(X, N, Z, show.progress=FALSE)

    # pdf('DENDRO.plot.pdf', width=8, height=8)
    # cluster = DENDRO.cluster(dist, label=colnames(Z), type='fan')
    # dev.off()

    suppressMessages(library(ggtree))
    suppressMessages(library(ggtreeExtra))
    suppressMessages(library(aplot))
    suppressMessages(library(ggplot2))
    suppressMessages(library(cowplot))

    {body1}

    info1 <- fread(file)
    info1 <- info1 %>% as.data.frame()
    tree <- hclust(dist, method='ward.D')
    
    # p <- ggtree(tree, layout='fan', size=0.5) %<+% info1 +
    #     geom_tippoint(mapping=aes(color=MPS), size=5) +
    #     scale_color_gradient2(low='#E41A1C', midpoint=0, mid='white', high='#377EB8')

    {body2}

    ggsave(p, filename='DENDRO.final.pdf', device='pdf',
        width={width}, height={height}, units='in', limitsize=FALSE)
    """
    robjects.r(cmd)

    tsc.ul.cleanup(tmpdir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: May 16, 2020
# Description: annotating mutations by ANNOVAR to find the missense ones
# =========================================================================================

from trisicell.ul._servers import *


def run08(config, afterok):
    def cmds():
        cmds = ""
        cmds += cmd([f"mkdir -p {config['outdir']}/_annotating"])
        cmds += cmd([f"mkdir -p {config['outdir']}/_annotating/annovar_pre"])
        cmds += cmd([f"mkdir -p {config['outdir']}/_annotating/annovar_post"])
        cmds += cmd([f"mkdir -p {config['outdir']}/_annotating/annovar_final"])
        cmds += cmd(
            [
                f"{config['ANNOVAR']}/convert2annovar.pl",
                f"--format vcf4",
                f"{config['outdir']}/_calling/jointcalls.g.vcf",
                f"--outfile {config['outdir']}/_annotating/annovar_pre/pre",
                f"--includeinfo",
                f"--allsample",
            ]
        )
        for file in [
            f"{config['outdir']}/_annotating/annovar_pre/pre.{s}.avinput"
            for s in config["samples"]
        ]:
            cmds += cmd(
                [
                    f"{config['ANNOVAR']}/annotate_variation.pl",
                    f"--outfile {config['outdir']}/_annotating/annovar_post/post.{os.path.basename(file)[4:]}",
                    f"--dbtype {config['dbtype']}",
                    f"--buildver {config['buildver']}",
                    f"{file}",
                    f"{config['ANNOVAR']}/{config['whichdb']}/",
                ]
            )
        for file in [
            f"{config['outdir']}/_annotating/annovar_post/post.{s}.avinput.exonic_variant_function"
            for s in config["samples"]
        ]:
            # cmds += cmd([
            # 'awk -F \'\\t\' \'$2!="synonymous SNV" {print $0}\'',
            #     file,
            #     '| awk -F \'\\t\' \'$2!="unknown" {split($3,a,":");',
            #     'print a[2]"_"a[1]"\\t"$4"\\t"$5"\\t"$7"\\t"$8"\\t"$2"\\t"$3}\'',
            #     f'> {config['outdir']}/_annotating/annovar_final/{os.path.basename(file)[5:-32]}.txt',
            # ])
            cmds += cmd(
                [
                    'awk -F \'\\t\' \'$2!="unknown" {split($3,a,":");',
                    'print a[2]"_"a[1]"\\t"$4"\\t"$5"\\t"$7"\\t"$8"\\t"$2"\\t"$3}\'',
                    file,
                    f"> {config['outdir']}/_annotating/annovar_final/{os.path.basename(file)[5:-32]}.txt",
                ]
            )
        cmds += cmd([f"echo Done!"], islast=True)
        return cmds

    df = pd.DataFrame()
    df["cmd"] = [cmds()]

    jobname = os.path.basename(__file__)[:-3]
    cmdmain = write_cmds_get_main(
        df,
        jobname,
        config["time08"],
        config["mem08"],
        "",
        1,
        config["email"],
        config["tmpdir"],
        afterok,
    )
    return cmdmain



import trisicell as tsc


def infer_cna(adata):
    try:
        import rpy2.robjects as robjects
        from rpy2.robjects.packages import importr
    except:
        raise SystemExit(
            "A problem was encountered importing `rpy2`. "
            "To run this `rpy2` and `R` need to be installed."
        )
    try:
        importr("infercna")
    except:
        raise SystemExit(
            "A problem was encountered importing `infercna` in R. "
            "To run this `infercna` needs to be installed in R. "
            "Use the following lines to installed them.\n\n"
            "devtools::install_github(c('jlaffy/scalop','jlaffy/infercna'))\n"
        )

    tsc.logg.info(f"running InferCNA")
    tmpdir = tsc.ul.tmpdir(prefix="trisicell.", suffix=".cn", dirname=".")

    cmd = f"""
    suppressMessages(library(infercna))
    suppressMessages(library(data.table))
    suppressMessages(library(ggplot2))
    useGenome('hg19')
    ref <- retrieveGenome()

    tpm <- fread('~/crc1/expr.tpm.tsv') %>% as.data.frame()
    row.names(tpm) <- tpm$V1
    tpm$V1 <- NULL
    tpm <- t(as.matrix(tpm))
    tpm <- log2(tpm/10 + 1)
    row.names(tpm) <- sapply(row.names(tpm),
                             function(x) strsplit(x, '_')[[1]][2],
                             USE.NAMES=FALSE)

    refCells <- list('normalcolon' = c('crc1_NC_502'))

    cna = infercna(m=tpm, refCells=refCells, n=5000, noise=0.1,
                   isLog=TRUE, verbose=FALSE)
    cnaM = cna[, !colnames(cna) %in% unlist(refCells)]
    obj = cnaPlot(cna = cnaM)

    ggsave(obj$p, filename='crc1.png', device='png',
           width=7.5, height=6, units='in', limitsize=FALSE)
    """
    robjects.r(cmd)

    tsc.ul.cleanup(tmpdir)



import csv
import glob
import gzip
import math
import os
import shutil
import subprocess

import click
import pandas as pd
import yaml
from natsort import natsorted

import trisicell as tsc

from .mcalling.s01indexing import run01
from .mcalling.s02mapping import run02
from .mcalling.s03indexing import run03
from .mcalling.s04mapping import run04
from .mcalling.s05calling import run05
from .mcalling.s06jointcalling import run06
from .mcalling.s07merging import run07
from .mcalling.s08annotating import run08
from .mcalling.s09expressing import run09
from .mcalling.s10velocitying import run10
from .mcalling.z01status import after01
from .mcalling.z02matrixing import after02
from .mcalling.z03clean import after03


def find_samples_readlength(config):
    def remove_pre_post(infile):
        if config["pairedend"]:
            return (
                os.path.basename(infile)
                .replace(config["infqpost1"], "")
                .replace(config["infqpost2"], "")
            )
        else:
            return os.path.basename(infile).replace(config["infqpost"], "")

    def get_readlength(filename):
        readlength = None
        with gzip.open(filename, "rb") as fin:
            i = 0
            for line in fin:
                readlength = len(line.strip()) - 1
                if i == 1:
                    break
                i += 1
        return readlength

    if os.path.exists(config["infq"]):
        infiles = glob.glob(f"{config['infq']}/*.fastq.gz")
        samples = natsorted(list(set(map(remove_pre_post, infiles))))
        readlength = get_readlength(infiles[0])
    else:
        infiles = glob.glob(f"{config['outdir']}/*/")
        samples = []
        for file in infiles:
            sample = file.split("/")[-2]
            if sample[0] != "_":
                samples.append(sample)
        samples = natsorted(samples)
        readlength = 73  # TODO: extract automatically
    return samples, readlength


@click.command(short_help="Mutation calling.")
@click.argument(
    "config_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.option(
    "--test",
    "-t",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="Is in test mode.",
)
@click.option(
    "--status",
    "-s",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="Check the status of runs.",
)
@click.option(
    "--build",
    "-b",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="Build the final matrices of genotype and expression.",
)
@click.option(
    "--clean",
    "-c",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="Cleaning the directory including removing intermediate files.",
)
def mcalling(config_file, test, status, build, clean):
    """Mutation calling pipeline.

    trisicell mcalling config.yml
    """

    with open(config_file) as fin:
        config = yaml.load(fin, Loader=yaml.FullLoader)

        samples, readlength = find_samples_readlength(config)
        config["samples"] = samples
        config["readlength"] = readlength

        tsc.logg.info(f"isRNA = {config['isrna']}")
        tsc.logg.info(config["samples"][:2], len(config["samples"]))
        tsc.logg.info(config["readlength"])

        tsc.ul.mkdir(config["tmpdir"] + "/cmd")
        tsc.ul.mkdir(config["tmpdir"] + "/log")
        tsc.ul.mkdir(config["outdir"])

        if config["buildver"] in ["hg19", "hg38"]:
            config["chroms"] = [f"chr{i}" for i in range(1, 23)] + ["chrX", "chrY"]
        elif config["buildver"] == "mm10":
            config["chroms"] = [f"chr{i}" for i in range(1, 20)] + ["chrX", "chrY"]

        if "infqpost1" in config and "infqpost2" in config:
            config["infqpost3"] = f"{config['infqpost1'].split('.')[0]}_val_1.fq.gz"
            config["infqpost4"] = f"{config['infqpost2'].split('.')[0]}_val_2.fq.gz"

        if "infqpost" in config:
            config["infqpost5"] = f"{config['infqpost']}_trimmed.fq.gz"

        #### STEPS
        if (not status) and (not build) and (not clean):
            if config["isrna"] == True:
                out01 = None
                if config["steps"]["s01indexing"]:
                    cmd01 = run01(config)
                    if not test:
                        out01 = subprocess.getoutput(cmd01)

                out02 = None
                if config["steps"]["s02mapping"]:
                    cmd02 = run02(config, out01)
                    if not test:
                        out02 = subprocess.getoutput(cmd02)

                out03 = None
                if config["steps"]["s03indexing"]:
                    cmd03 = run03(config, out02)
                    if not test:
                        out03 = subprocess.getoutput(cmd03)

                out04 = None
                if config["steps"]["s04mapping"]:
                    cmd04 = run04(config, out03)
                    if not test:
                        out04 = subprocess.getoutput(cmd04)

                out05 = None
                if config["steps"]["s05calling"]:
                    cmd05 = run05(config, out04)
                    if not test:
                        out05 = subprocess.getoutput(cmd05)

                out06 = None
                if config["steps"]["s06jointcalling"]:
                    cmd06 = run06(config, out05)
                    if not test:
                        out06 = subprocess.getoutput(cmd06)

                out07 = None
                if config["steps"]["s07merging"]:
                    cmd07 = run07(config, out06)
                    if not test:
                        out07 = subprocess.getoutput(cmd07)

                out08 = None
                if config["steps"]["s08annotating"]:
                    cmd08 = run08(config, out07)
                    if not test:
                        out08 = subprocess.getoutput(cmd08)

                out09 = None
                if config["steps"]["s09expressing"]:
                    cmd09 = run09(config, out05)
                    if not test:
                        out09 = subprocess.getoutput(cmd09)

                out10 = None
                if config["steps"]["s10velocitying"]:
                    cmd10 = run10(config, out04)
                    if not test:
                        out10 = subprocess.getoutput(cmd10)
            else:
                out02 = None
                if config["steps"]["s02mapping"]:
                    cmd02 = run02(config, None)
                    if not test:
                        out02 = subprocess.getoutput(cmd02)

                out04 = None
                if config["steps"]["s04mapping"]:
                    cmd04 = run04(config, out02)
                    if not test:
                        out04 = subprocess.getoutput(cmd04)

                out05 = None
                if config["steps"]["s05calling"]:
                    cmd05 = run05(config, out04)
                    if not test:
                        out05 = subprocess.getoutput(cmd05)

                out06 = None
                if config["steps"]["s06jointcalling"]:
                    cmd06 = run06(config, out05)
                    if not test:
                        out06 = subprocess.getoutput(cmd06)

                out07 = None
                if config["steps"]["s07merging"]:
                    cmd07 = run07(config, out06)
                    if not test:
                        out07 = subprocess.getoutput(cmd07)

                out08 = None
                if config["steps"]["s08annotating"]:
                    cmd08 = run08(config, out07)
                    if not test:
                        out08 = subprocess.getoutput(cmd08)
        else:
            if status:
                after01(config)
            if build:
                after02(config)
            if clean:
                after03(config)

    return None



import math

import numpy as np

import trisicell as tsc


def bifiltering(df, cellr, mutr, time_limit=3600):
    """Bi-filtering to find maximally inforemed submatrix.

    This function runs an ILP to find maximally inforemed submatrix
    where the number of mutant genotypes is maximized.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        The input noisy genotype matrix where entries are 0,1 and 3.
    cellr : :obj:`float`
        ratio for picking how many cells.
    mutr : :obj:`float`
        ratio for picking how many mutations.
    time_limit : :obj:`int`, optional
        Time limit for the ILP solver, by default 3600.

    Returns
    -------
    :class:`pandas.DataFrame`
        The output gentoype submatrix.
    """

    gp, gp_is_not_imported = tsc.ul.import_gurobi()
    if gp_is_not_imported:
        raise RuntimeError("Unable to import a package!")

    M = df.values.copy()
    #### only mutant
    # M[M != 1] = -1

    #### mutant and reference
    M[M != 3] = 1
    M[M == 3] = -1
    n, m = M.shape
    n_cells, n_sites = math.floor(cellr * n), math.floor(mutr * m)

    model = gp.Model("bi-filtering")
    model.Params.OutputFlag = 0
    model.Params.LogFile = ""
    model.Params.Threads = 1
    model.Params.TimeLimit = time_limit

    tsc.logg.info("1. declaring variables", time=True)
    A, C, S = {}, {}, {}
    for i in range(n):
        A[i] = {}
        for j in range(m):
            A[i][j] = model.addVar(vtype=gp.GRB.BINARY)
    for i in range(n):
        C[i] = model.addVar(vtype=gp.GRB.BINARY)
        for j in range(m):
            S[j] = model.addVar(vtype=gp.GRB.BINARY)
    # A = model.addMVar((n, m), vtype=gp.GRB.BINARY)
    # C = model.addMVar(n, vtype=gp.GRB.BINARY)
    # S = model.addMVar(m, vtype=gp.GRB.BINARY)

    tsc.logg.info("2. adding constraints", time=True)
    for i in range(n):
        for j in range(m):
            model.addConstr(A[i][j] <= C[i])
            model.addConstr(A[i][j] <= S[j])
            model.addConstr(C[i] + S[j] - 1 <= A[i][j])
    # for i in range(n):
    #     for j in range(m):
    #         model.addConstr(A[i, j] <= C[i])
    #         model.addConstr(A[i, j] <= S[j])
    #         model.addConstr(C[i] + S[j] - 1 <= A[i, j])

    model.addConstr(gp.quicksum(C[i] for i in range(n)) == n_cells)
    model.addConstr(gp.quicksum(S[j] for j in range(m)) == n_sites)
    # model.addConstr(sum(C[i] for i in range(n)) == n_cells)
    # model.addConstr(sum(S[j] for j in range(m)) == n_sites)

    model.update()

    tsc.logg.info("3. adding objective", time=True)
    model.setObjective(
        gp.quicksum(A[i][j] * M[i, j] for i in range(n) for j in range(m)),
        gp.GRB.MAXIMIZE,
    )

    tsc.logg.info("4. start solving", time=True)
    model.optimize()

    a = np.array([i for i in range(n) if C[i].X > 0])
    b = np.array([j for j in range(m) if S[j].X > 0])
    cells = df.index[a]
    sites = df.columns[b]

    return df.loc[cells, sites].copy()



import os

import pytest
from _helpers import *
from click.testing import CliRunner

import trisicell as tsc
from trisicell.commands.trisicell import cli


class TestCommands:
    def test_scistree(self):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "scistree",
                f"{tsc.ul.get_file('trisicell.datasets/test/test.SC')}",
                "0.0000001",
                "0.1",
            ],
        )
        assert result.exit_code == 0

    def test_scite(self):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "scite",
                f"{tsc.ul.get_file('trisicell.datasets/test/test.SC')}",
                "0.0000001",
                "0.1",
                "-r 3",
                "-l 1000",
            ],
        )
        assert result.exit_code == 0

    def test_phiscsb(self):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "phiscsb",
                f"{tsc.ul.get_file('trisicell.datasets/test/test.SC')}",
                "0.0000001",
                "0.1",
            ],
        )
        assert result.exit_code == 0

    def test_consensus(self):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "consensus",
                f"{tsc.ul.get_file('trisicell.datasets/test/consensus/biorxiv.fig3b.CFMatrix')}",
                f"{tsc.ul.get_file('trisicell.datasets/test/consensus/biorxiv.figs18a.CFMatrix')}",
                f"{tsc.ul.get_file('trisicell.datasets/test/consensus.CFMatrix')}",
            ],
        )
        assert result.exit_code == 0


@pytest.fixture(scope="session", autouse=True)
def cleanup(request):
    def remove_test_dir():
        tsc.ul.remove(tsc.ul.get_file("trisicell.datasets/test/test.scistree.CFMatrix"))
        tsc.ul.remove(tsc.ul.get_file("trisicell.datasets/test/test.scistree.log"))
        tsc.ul.remove(tsc.ul.get_file("trisicell.datasets/test/test.scite.CFMatrix"))
        tsc.ul.remove(tsc.ul.get_file("trisicell.datasets/test/test.scite.log"))
        tsc.ul.remove(tsc.ul.get_file("trisicell.datasets/test/test.phiscsb.CFMatrix"))
        tsc.ul.remove(tsc.ul.get_file("trisicell.datasets/test/test.phiscsb.log"))
        tsc.ul.remove(tsc.ul.get_file("trisicell.datasets/test/consensus.CFMatrix"))

    request.addfinalizer(remove_test_dir)



import contextlib
import functools
import multiprocessing
import os
import time

import joblib
import networkx as nx
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from networkx.drawing.nx_agraph import graphviz_layout, to_agraph
from tqdm import tqdm

import trisicell as tsc


def with_timeout(timeout):
    def decorator(decorated):
        @functools.wraps(decorated)
        def inner(*args, **kwargs):
            pool = multiprocessing.pool.ThreadPool(1)
            async_result = pool.apply_async(decorated, args, kwargs)
            try:
                return async_result.get(timeout)
            except multiprocessing.TimeoutError:
                return None

        return inner

    return decorator


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


def subsampling(
    df_input,
    alpha,
    beta,
    solver,
    sample_on,
    sample_size,
    n_samples,
    begin_sample,
    n_jobs,
    time_out,
    tmpdir,
    disable_tqdm,
):

    # tsc.logg.info(
    #     f"running Booster with alpha={alpha}, beta={beta}, solver={solver},"
    #     f" sample_on={sample_on}, sample_size={sample_size}, n_samples={n_samples},"
    #     f" n_jobs={n_jobs}, time_out={time_out}"
    # )
    # tsc.logg.info(f"id,m,i0,i1,i3,o0,o1,f01,f10,f30,f31,r")

    # graph = nx.DiGraph()
    # graph.add_nodes_from(df_input.columns)

    # def add_edge(muty, mutx):
    #     if graph.has_edge(muty, mutx):
    #         graph[muty][mutx]["label"] += 1
    #     else:
    #         graph.add_edge(muty, mutx, label=1)

    # def remove_edge(muty, mutx):
    #     if graph.has_edge(muty, mutx):
    #         graph[muty][mutx]["label"] -= 1

    @with_timeout(time_out)
    def run(i):
        if sample_on == "muts":
            dfn = df_input.sample(n=sample_size, replace=False, axis=1)
        elif sample_on == "cells":
            dfn = df_input.sample(n=sample_size, replace=False, axis=0)

        if solver.lower() == "phiscs":
            dfo = tsc.tl.phiscsb(dfn, alpha, beta, experiment=True)
            dfo.to_csv(f"{tmpdir}/{i}.CFMatrix", sep="\t")
        elif solver.lower() == "scite":
            dfo, _, _, _ = tsc.tl.scite(
                dfn, alpha, beta, time_out, 1, False, experiment=True
            )
            dfo.to_csv(f"{tmpdir}/{i}.CFMatrix", sep="\t")
        elif solver.lower() == "scistree":
            x = ((dfn == 1).sum() >= 2) & ((dfn != 3).sum() >= 0.2 * dfn.shape[0])
            dfn = dfn[dfn.columns[x]]
            if dfn.shape[1] < 2:
                return None
            dfo, run_time = tsc.tl.scistree(dfn, alpha, beta, False, experiment=True)
            i0 = np.sum(dfn.values == 0)
            i1 = np.sum(dfn.values == 1)
            i3 = np.sum(dfn.values == 3)
            o0 = np.sum(dfo.values == 0)
            o1 = np.sum(dfo.values == 1)
            f01, f10, f30, f31 = tsc.ul.count_flips(dfn.values, dfo.values)
            # tsc.logg.info(
            #     f"{i},{dfn.shape[1]},{i0},{i1},{i3},{o0},{o1},{f01},{f10},{f30},{f31},{run_time:.1f}"
            # )
            dfo.to_csv(f"{tmpdir}/{i}.CFMatrix", sep="\t")

    s_time = time.time()
    with tqdm_joblib(
        tqdm(
            ascii=True,
            ncols=100,
            desc="SUBSAMPLING ",
            total=n_samples,
            position=0,
            disable=disable_tqdm,
        )
    ) as progress_bar:
        output = Parallel(n_jobs=n_jobs)(
            delayed(run)(i) for i in range(begin_sample, begin_sample + n_samples)
        )
    e_time = time.time()
    running_time = e_time - s_time

    # muts = np.append(["ROOT"], df_input.columns)
    # A = pd.DataFrame(0, index=muts, columns=muts)
    # D = pd.DataFrame(0, index=muts, columns=muts)
    # for i in range(n_samples):
    #     try:
    #         dfo = pd.read_csv(f"{tmpdir}/{i}.CFMatrix", sep="\t", index_col=0)
    #     except:
    #         continue
    #     for i in range(dfo.shape[1]):
    #         for j in range(dfo.shape[1]):
    #             if i != j:
    #                 mutx = dfo.columns[i]
    #                 muty = dfo.columns[j]
    #                 if (dfo[mutx] == dfo[muty]).all():
    #                     add_edge(muty, mutx)
    #                     A.loc[muty, mutx] += 1
    #                 elif (dfo[mutx] <= dfo[muty]).all():
    #                     add_edge(muty, mutx)
    #                     A.loc[muty, mutx] += 1
    #                 else:
    #                     D.loc[mutx, muty] += 1

    # for i in range(n_samples):
    #     try:
    #         dfo = pd.read_csv(f"{tmpdir}/{i}.CFMatrix", sep="\t", index_col=0)
    #     except:
    #         continue
    #     for i in range(dfo.shape[1]):
    #         for j in range(dfo.shape[1]):
    #             if i != j:
    #                 mutx = dfo.columns[i]
    #                 muty = dfo.columns[j]
    #                 if (dfo[mutx] == dfo[muty]).all():
    #                     pass
    #                 elif (dfo[mutx] <= dfo[muty]).all():
    #                     pass
    #                 else:
    #                     remove_edge(muty, mutx)
    #                     remove_edge(mutx, muty)

    # A.loc["ROOT", :] = np.inf
    # D.loc["ROOT", :] = 0
    # D.loc[:, "ROOT"] = 0
    # A.loc[:, "ROOT"] = 0

    # graph.add_nodes_from(["ROOT"])
    # for mut in df_input.columns:
    #     add_edge("ROOT", mut)

    # T = nx.algorithms.tree.maximum_spanning_arborescence(graph, attr="label")
    # mygraph = to_agraph(T)
    # mygraph.layout(prog="dot")
    # mygraph.draw(f"{tmpdir}/_tree.png")

    # tsc.ul.stat(df_input, df_output, alpha, beta, running_time)



import trisicell as tsc


def import_gurobi():
    try:
        import gurobipy as gp

        # m = gp.Model("MIP")
        # m.Params.OutputFlag = 0
        # m.Params.LogFile = ""
        # m.Params.Threads = 1
        # x = m.addVar(vtype=gp.GRB.BINARY, name="x")
        # y = m.addVar(vtype=gp.GRB.BINARY, name="y")
        # z = m.addVar(vtype=gp.GRB.BINARY, name="z")
        # m.setObjective(x + y + 2 * z, gp.GRB.MAXIMIZE)
        # m.addConstr(x + 2 * y + 3 * z <= 4, "c0")
        # m.addConstr(x + y >= 1, "c1")
        # m.optimize()
        return gp, False
    except:
        tsc.logg.error(
            "Unable to import `gurobipy`!",
            "Make sure `Gurobi` is already installed in your system.",
            "Then install `gurobipy`.",
        )
        return None, True


def import_mpi4py():
    try:
        import mpi4py

        return mpi4py, False
    except:
        tsc.logg.error(
            "Unable to import `mpi4py`!",
            "Make sure `mpi/mpich-x86_64` is already installed in your system.",
            "Then install `mpi4py`.",
        )
        return None, True


def import_rpy2(name="base", how=""):
    "Unable to import `rpy2`, install it first as `pip install rpy2` version `>=3.3.0`."
    try:
        from rpy2.robjects import r
        from rpy2.robjects.packages import PackageNotInstalledError, importr

    except ImportError:
        tsc.logg.error(
            "Unable to import `rpy2`, install it first as `pip install rpy2` version `>=3.3.0`."
        )
        return None, True

    try:
        _r_lib = importr(name)
        return _r_lib, False
    except PackageNotInstalledError:
        tsc.logg.error(f"Install R library `{name!r}` first." f"{how}")
        return None, True


def import_graphviz():
    # TODO: implement
    return None, None



import os

import numpy as np
import pandas as pd

import trisicell as tsc


def siclonefit(df_input, alpha, beta, n_iters, save_inter=False):
    tsc.logg.info(
        f"running SiCloneFit with alpha={alpha}, beta={beta}, n_iters={n_iters}"
    )
    tmpdir = tsc.ul.tmpdir(prefix="trisicell.", suffix=".siclonefit", dirname=".")

    df_input.T.reset_index(drop=True).to_csv(
        f"{tmpdir}/siclonefit.input", sep=" ", header=None
    )
    with open(f"{tmpdir}/siclonefit.cellnames", "w") as fout:
        fout.write(" ".join(df_input.index))
    with open(f"{tmpdir}/siclonefit.genenames", "w") as fout:
        fout.write(" ".join(df_input.columns))
    I = df_input.values

    siclonefit = tsc.ul.get_file("trisicell.external/bin/SiCloneFiTComplete.jar")
    cmd = (
        f"java -jar {siclonefit} "
        f"-m {df_input.shape[0]} "
        f"-n {df_input.shape[1]} "
        f"-ipMat {tmpdir}/siclonefit.input "
        f"-fp {alpha} "
        f"-fn {beta} "
        "-df 0 "
        f"-missing {np.sum(I == 3)/(I.size)} "
        "-f 3 "
        "-recurProb 0 "
        "-delProb 0 "
        "-LOHProb 0 "
        f"-iter {n_iters} "
        f"-cellNames {tmpdir}/siclonefit.cellnames "
        f"-geneNames {tmpdir}/siclonefit.genenames "
        f"-outDir {tmpdir} > {tmpdir}/siclonefit.log"
    )
    # check the following parameters
    # -burnin
    # -printIter
    # -treeIter
    # -doublet
    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    df = pd.read_csv(
        f"{tmpdir}/20p_missing_samples/best/best_MAP_predicted_genotype.txt",
        sep=" ",
        header=None,
        index_col=0,
    ).T
    df_output = pd.DataFrame(df.values)
    df_output.columns = df_input.columns
    df_output.index = df_input.index
    df_output.index.name = "cellIDxmutID"

    if not save_inter:
        tsc.ul.cleanup(tmpdir)

    tsc.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



import pandas as pd
import yaml

import trisicell as tsc


def get_markers():
    markers = pd.read_csv(tsc.ul.get_file("trisicell.datasets/data/markers.csv"))
    return markers


def get_signatures(kind):
    signature_file = ""
    if kind == "mm10":
        signature_file = tsc.ul.get_file("trisicell.datasets/data/signatures_mm10.yml")
    elif kind == "hg19":
        signature_file = tsc.ul.get_file("trisicell.datasets/data/signatures_hg19.yml")
    with open(signature_file) as fin:
        signatures = yaml.load(fin, Loader=yaml.FullLoader)
    return signatures



import pytest

import trisicell as tsc

skip_gurobi = pytest.mark.skipif(
    tsc.ul.import_gurobi()[1], reason="Unable to import `Gurobi`!"
)
skip_mpi4py = pytest.mark.skipif(
    tsc.ul.import_mpi4py()[1], reason="Unable to import `MPI`!"
)
skip_rpy2 = pytest.mark.skipif(
    tsc.ul.import_rpy2()[1], reason="Unable to import `rpy2`!"
)
skip_graphviz = pytest.mark.skipif(
    tsc.ul.import_graphviz()[1], reason="Unable to import `Graphviz`!"
)



import pickle
import time
from decimal import *

import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import pairwise_distances
from tqdm import tqdm

from ._clt_sampler import draw_sample_clt


def cell_lineage_tree_prob(P, subtrees):
    """
    Calculate Prob_{A\sim P}[A\in T]
    O(n m^2)
    :param P:
    :param subtrees: cell lineage tree
    :return: Probability of this tree in the original distribution( based on P)
    """
    return_value = Decimal(1)
    for j in range(P.shape[1]):
        temp = Decimal(0)
        col = P[:, j]
        for v in subtrees:
            temp += Decimal(np.product(col * v + (1 - col) * (1 - v)))
        return_value *= temp
    return return_value


def pf_cond_on_one_tree(P, subtrees, cond_c, cond_m):
    """
    Prob_{A\sim P}[\subtree(c, R, A)\cap A\in G| A\in T]
    O(n^2)
    :param P:
    :param subtrees: cell lineage tree  n x (2n+1)
    :param cond_c: set of cells
    :param cond_m: one mutation
    :return: conditioned on the given tree what is probability of the given partition based on P
        numerator, denominator are returned separately here
    """
    denominator = Decimal(0)
    numerator = Decimal(0)
    col = P[:, cond_m]
    for v in subtrees:
        prob = Decimal(np.product(col * v + (1 - col) * (1 - v)))
        denominator += prob
        if np.array_equal(v, cond_c):
            numerator = prob
    return numerator, denominator


def get_samples(P, n_samples, disable_tqdm=True):
    """
    N_s *
    :param P:
    :param n_samples:
    :return: `n_samples` of cell lineage trees. The function returns three lists with this length:
        edges_list: for each sample: (n-1) wedges. Each wedge [parent, left_child, right_child]
        subtrees_list: for each sample: 2n+1 number of cell subset masks, i.e., potential column.
        tree_our_prob_list: for each sample: the probability of us sampling the tree. Prob_{T\sim E}[T]
    """
    P = P.astype(np.float128)
    edges_list = []
    subtrees_list = []
    tree_our_prob_list = []
    for _ in tqdm(
        range(n_samples), ascii=True, ncols=100, desc=f"Sampling", disable=disable_tqdm
    ):
        edges, subtrees, prior_prob = draw_sample_clt(P, False, c=1, coef=10)
        edges_list.append(edges)
        subtrees_list.append(subtrees)
        tree_our_prob_list.append(prior_prob)
    return edges_list, subtrees_list, tree_our_prob_list


def get_samples_info(P, my_cell, my_mut, n_samples, subtrees_list=None):
    """
    Runs some processes on the given samples (if not given first gets the samples) and returns some raw data.
    If sampling was done internally, it outputs the full sample information

    If not given_samples:
        O(N * (T_nj * T_obj))
        T_obj = n m^2
        T_nj = n^2 m^2
    Else:
        O(N * n^2)
    :param P: A probability matrix:
            where P_{i,j} is the probablity of i\th cell having j\th mutation in the latent original matrix.
            In other words, if O is the original matrix and I is the measured (observed) matrix then,
            P_{i,j} = P[O_{i,j}=1|<I_{i,j}, \alpha, \beta>].
    :param my_cell:
    :param my_mut:
    :param n_samples:
    :param subtrees_list: presampled trees. If None new samples will be drawen
    :return:
        if given_samples:   (i.e., subtrees_list is not None)
            pf_cond_list, tree_origin_prob_list
        else:               (i.e., subtrees_list is None)
            pf_cond_list, tree_origin_prob_list, edges_list, subtrees_list, tree_our_prob_list
    """
    given_samples = subtrees_list is not None
    if not given_samples:
        edges_list, subtrees_list, tree_our_prob_list = get_samples(P, n_samples)

    pf_cond_list = []
    tree_origin_prob_list = []
    cond_c = np.zeros(P.shape[0], dtype=np.int8)
    cond_c[my_cell] = 1

    for i in tqdm(
        range(n_samples), ascii=True, ncols=100, desc=f"Sampling", disable=True
    ):
        numerator, denominator = pf_cond_on_one_tree(
            P, subtrees_list[i], cond_c=cond_c, cond_m=my_mut
        )
        pf_cond_list.append(numerator / denominator)

        origin_prob = cell_lineage_tree_prob(P, subtrees_list[i])
        tree_origin_prob_list.append(origin_prob)

    if given_samples:
        return pf_cond_list, tree_origin_prob_list
    else:
        return (
            pf_cond_list,
            tree_origin_prob_list,
            edges_list,
            subtrees_list,
            tree_our_prob_list,
        )


def process_samples(
    pf_cond_list, tree_origin_prob_list, tree_our_prob_list, n_batches=None
):
    """
    Combines the data corresponding to the sampled trees and output the partition function estimate
    according to the formula in the paper. One can see this as just a simple weighted average. Features:
        1. Debiasing is implemented through weights
            (in favour of original probability and against probability of our sampling
        2. Normalization: instead of calculating exact denominator given in the paper. We divide the value
            by summation of the weights.
    :param pf_cond_list:
    :param tree_origin_prob_list:
    :param tree_our_prob_list:
    :param n_batches TODO
    :return: just one probability through weighted average
    """
    n_samples = len(pf_cond_list)
    assert n_samples == len(tree_origin_prob_list)
    assert n_samples == len(tree_our_prob_list)
    estimates = list()
    n_batches_internal = n_batches if n_batches is not None else 1
    interval_len = n_samples // n_batches_internal
    for batch_ind in range(n_batches_internal):
        numerator = 0
        denominator = 0
        samples_interval_start = batch_ind * interval_len
        samples_interval_end = min((batch_ind + 1) * interval_len, n_samples)
        assert samples_interval_start < samples_interval_end
        for i in range(samples_interval_start, samples_interval_end):
            numerator += (
                pf_cond_list[i] * tree_origin_prob_list[i] / tree_our_prob_list[i]
            )
            denominator += tree_origin_prob_list[i] / tree_our_prob_list[i]
        estimates.append(numerator / denominator)
    assert len(estimates) >= 1
    if n_batches is None:
        return estimates[0]
    else:
        return estimates


def count_disticnt_matrices(x_list):
    """
    Given a list of matrices, returns the number of distinct ones.
    :param x_list:
    :return:
    """
    hashes = set()
    for x in x_list:
        x = np.matrix(x)
        x.flags.writeable = False
        hashes.add(hash(x.tostring()))
    return len(hashes)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: May 16, 2020
# Description: calling mutations by GATK HaplotypeCaller
# =========================================================================================

from trisicell.ul._servers import *


def run05(config, afterok):
    def cmds(sample):
        cmds = ""
        cmds += cmd([f"mkdir -p {config['outdir']}/{sample}"])
        if len(config["normals"]) > 0:
            if sample not in config["normals"]:
                cmds += cmd(
                    [
                        f"{config['java05']} {config['GATK']}",
                        f"-T MuTect2",
                        f"-R {config['ref']}",
                        f"-I:tumor {config['outdir']}/{sample}/output.bam",
                        f"-I:normal {config['outdir']}/{config['normals'][0]}/output.bam",
                        f"-o {config['outdir']}/{sample}/MuTect2.g.vcf",
                        f"-dontUseSoftClippedBases",
                        f"-stand_call_conf 20",
                        f"--dbsnp {config['hgDBSNPs138']}"
                        if buildver == "hg19"
                        else f"--dbsnp {config['mmDBSNPs142']}",
                        f"-ERC GVCF",
                    ]
                )
        else:
            cmds += cmd(
                [
                    f"{config['java05']} {config['GATK']}",
                    f"-T HaplotypeCaller",
                    f"-R {config['ref']}",
                    f"-I {config['outdir']}/{sample}/output.bam",
                    f"-o {config['outdir']}/{sample}/HaplotypeCaller.g.vcf",
                    f"-dontUseSoftClippedBases",
                    f"-stand_call_conf 20",
                    f"--dbsnp {config['hgDBSNPs138']}"
                    if config["buildver"] == "hg19"
                    else f"--dbsnp {config['mmDBSNPs142']}",
                    f"-ERC GVCF",
                ]
            )
        cmds += cmd([f"echo Done!"], islast=True)
        return cmds

    df = pd.DataFrame()
    df["sample"] = config["samples"]
    df["cmd"] = df.apply(lambda x: cmds(x["sample"]), axis=1)

    jobname = os.path.basename(__file__)[:-3]
    cmdmain = write_cmds_get_main(
        df,
        jobname,
        config["time05"],
        config["mem05"],
        "python/3.7",
        1,
        config["email"],
        config["tmpdir"],
        afterok,
    )
    return cmdmain



import pytest
from _helpers import *

import trisicell as tsc


class TestPreProcessing:
    @skip_gurobi
    def test_bifiltering(self):
        df_in = tsc.datasets.test()
        df_filtered = tsc.pp.bifiltering(df_in, 0.5, 0.2)
        assert df_filtered.shape == (10, 4)



import os

import click

import trisicell as tsc


@click.command(short_help="Run Booster.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--solver",
    "-s",
    default="scite",
    type=click.Choice(["scite", "phiscs", "scistree"]),
    show_default=True,
    help="Solver of the booster.",
)
@click.option(
    "--sample_on",
    "-so",
    default="muts",
    type=click.Choice(["muts", "cells"]),
    show_default=True,
    help="Sampling on `muts` or `cells`.",
)
@click.option(
    "--sample_size",
    "-ss",
    default=10,
    type=int,
    show_default=True,
    help="The size of samples i.e. the number of muts or cells.",
)
@click.option(
    "--n_samples",
    "-ns",
    default=10,
    type=int,
    show_default=True,
    help="The number of subsamples.",
)
@click.option(
    "--begin_sample",
    "-bs",
    default=0,
    type=int,
    show_default=True,
    help="ID of the start subsample name.",
)
@click.option(
    "--n_jobs",
    "-nj",
    default=0,
    type=int,
    show_default=True,
    help="Number of parallel jobs to do subsampleing.",
)
@click.option(
    "--time_out",
    "-to",
    default=120,
    type=int,
    show_default=True,
    help="""Timeout of solving allowance
    (if solver is SCITE this parameter is the number of iterations)""",
)
@click.option(
    "--disable_tqdm",
    "-dt",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="Disable showing the tqdm progress.",
)
@click.option(
    "--weight",
    "-w",
    default=50,
    type=int,
    show_default=True,
    help="""Weight for how many subsamples to be used in dependencies calculation.""",
)
@click.option(
    "--no_subsampling",
    "-nss",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="No running of subsampling.",
)
@click.option(
    "--no_dependencies",
    "-nd",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="No running of subsampling.",
)
def booster(
    genotype_file,
    alpha,
    beta,
    solver,
    sample_on,
    sample_size,
    n_samples,
    begin_sample,
    n_jobs,
    time_out,
    disable_tqdm,
    weight,
    no_subsampling,
    no_dependencies,
):
    """Divide and Conquer

    trisicell booster input.SC 0.001 0.1
    -s scite -ss 40 -ns 10 -bs 0 -nj 5 -to 1000 -nd/-nss
    """

    dirbase = tsc.ul.dirbase(genotype_file)
    dirname, basename = tsc.ul.dir_base(genotype_file)

    tsc.settings.verbosity = "info"
    # tsc.settings.logfile = f'{dirbase}.booster.log'

    df_in = tsc.io.read(genotype_file)
    df_out = tsc.tl.booster(
        df_in,
        alpha=alpha,
        beta=beta,
        solver=solver,
        sample_on=sample_on,
        sample_size=sample_size,
        n_samples=n_samples,
        begin_sample=begin_sample,
        n_jobs=n_jobs,
        time_out=time_out,
        save_inter=True,
        dir_inter=".",
        base_inter=basename,
        disable_tqdm=disable_tqdm,
        weight=weight,
        no_subsampling=no_subsampling,
        no_dependencies=no_dependencies,
    )
    # tsc.io.write(df_out, f'{dirbase}.booster.CFMatrix')

    return None



import numba
import numpy as np
from sklearn.metrics import pairwise_distances


def _l1_ignore_na(a, b):
    a[a == 3] = np.nan
    b[b == 3] = np.nan
    return np.nanmean(np.abs(a - b))


def calc_dist_by_l1_ignore_na(I):
    return pairwise_distances(I, metric=_l1_ignore_na)


# https://gist.github.com/FedericoV/0e7d6d8c8794a99a7a42
@numba.jit(nopython=True)
def _cosine_ignore_na(u, v):
    m = u.shape[0]
    udotv = 0
    u_norm = 0
    v_norm = 0
    for i in range(m):
        if (np.isnan(u[i])) or (np.isnan(v[i])):
            continue
        udotv += u[i] * v[i]
        u_norm += u[i] * u[i]
        v_norm += v[i] * v[i]
    u_norm = np.sqrt(u_norm)
    v_norm = np.sqrt(v_norm)
    if (u_norm == 0) or (v_norm == 0):
        ratio = 1.0
    else:
        ratio = 1 - udotv / (u_norm * v_norm)
    if ratio < 0:
        return 0
    return ratio


def calc_dist_by_cosine_ignore_na(I):
    return pairwise_distances(I, metric=_cosine_ignore_na)


def calc_dist_by_dendro(T, V, I):
    from betabinom import pmf_BetaBinomial
    from scipy import stats
    from scipy.stats import binom

    PROB_SEQ_ERROR = 0.001

    def logSum_1(x, y):
        big = np.copy(x)
        big[x < y] = y[x < y]
        small = np.copy(x)
        small[x >= y] = y[x >= y]
        tmp = big + np.log(1 + np.exp(small - big))
        # tmp[np.bitwise_and(x==-np.inf, y==-np.inf)] = -np.inf
        # tmp[np.bitwise_and(x==np.inf, y==np.inf)] = np.inf
        return tmp

    D = np.divide(V, T)

    Mu = np.nanmean(D, axis=0)
    Var = np.nanvar(D, axis=0, ddof=1)
    a = ((1 - Mu) * Mu / Var - 1) * Mu
    b = ((1 - Mu) * Mu / Var - 1) * (1 - Mu)
    bad_muts = (
        (a <= 0) | (b <= 0) | np.isnan(a) | np.isnan(b) | np.isinf(a) | np.isinf(b)
    )
    V = V[:, ~bad_muts]
    T = T[:, ~bad_muts]
    D = D[:, ~bad_muts]
    I = I[:, ~bad_muts]
    a = a[~bad_muts]
    b = b[~bad_muts]

    lPz0 = np.zeros(T.shape, dtype=np.float64)
    lPz1 = np.zeros(T.shape, dtype=np.float64)
    for i in range(T.shape[0]):
        for j in range(T.shape[1]):
            if T[i, j] != 0:
                lPz0[i, j] = np.log(binom.pmf(V[i, j], T[i, j], PROB_SEQ_ERROR))
                lPz1[i, j] = np.log(pmf_BetaBinomial(V[i, j], T[i, j], a[j], b[j]))

    Pg = np.sum(I == 1, axis=0) / I.shape[0]
    lPg = np.log(Pg)
    l1Pg = np.log(1 - Pg)
    lupiall = logSum_1(lPz0 + l1Pg, lPz1 + lPg)

    dist = np.zeros((T.shape[0], T.shape[0]), dtype=np.float64)
    for i in range(T.shape[0]):
        ldowni = logSum_1(lPz0[i, :] + lPz0 + l1Pg, lPz1[i, :] + lPz1 + lPg)
        lupi = logSum_1(lupiall[i, :] + lupiall, ldowni)
        dist[i, :] = np.sum(lupi - ldowni, axis=1)

    dist = dist - np.min(dist) + 1
    return dist, bad_muts



import copy
import os
import time
from decimal import *

import Bio.Phylo as bp
import numpy as np
import pandas as pd
from Bio.Phylo import BaseTree
from Bio.Phylo.TreeConstruction import DistanceMatrix, DistanceTreeConstructor

import trisicell as tsc


def scistree(df_input, alpha, beta, save_inter=False, experiment=False):
    """Solving using ScisTree.

    Accurate and efficient cell lineage tree inference from noisy
    single cell data: the maximum likelihood perfect phylogeny approach :cite:`ScisTree`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    save_inter : :obj:`bool`, optional
        Save input/output of the ScisTree format, by default False
    experiment : :obj:`bool`, optional
        Is in the experiment mode (the log won't be shown), by default False

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """

    if not experiment:
        tsc.logg.info(f"running ScisTree with alpha={alpha}, beta={beta}")
    tmpdir = tsc.ul.tmpdir(prefix="trisicell.", suffix=".scistree", dirname=".")
    cells = df_input.index
    snvs = df_input.columns
    matrix_input = df_input.values
    df = df_input.transpose()

    df = df.replace(3, 0.5)
    df = df.replace(0, 1 - beta)
    df = df.replace(1, alpha)

    file1 = f"{tmpdir}/scistree.input"
    df.index.name = f"HAPLOID {df.shape[0]} {df.shape[1]}"
    df.to_csv(file1, sep=" ")
    with open(file1, "r") as ifile:
        data = ifile.read()
    with open(file1, "w") as ofile:
        data = data.replace('"', "")
        ofile.write(data)

    scistree = tsc.ul.get_file("trisicell.external/bin/scistree")
    cmd = (
        f"{scistree} "
        "-v "
        "-d 0 "
        "-e "
        f"-o {tmpdir}/scistree.gml "
        f"{tmpdir}/scistree.input > {tmpdir}/scistree.output"
    )
    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    data = []
    mut_tree = ""
    cell_tree = ""
    detail = {"cost": "\n"}
    with open(f"{tmpdir}/scistree.output") as infile:
        now_store = False
        for line in infile:
            line = line.strip()
            if "Mutation tree:" in line:
                mut_tree = line.split(":")[1].replace(" ", "")
                mut_tree = mut_tree.replace("#", "")
            if "Constructed single cell phylogeny:" in line:
                cell_tree = line.split(":")[1].replace(" ", "")
            if "Imputed genotypes:" in line:
                now_store = True
            if line[:4] == "Site" and now_store:
                line = "".join(line.split(":")[1])
                line = line.replace("\t", "")
                data.append([int(x) for x in line.split(" ")])
            if "current cost: " in line:
                cost = float(line.split("current cost: ")[1].split(", opt tree: ")[0])
                detail["cost"] += f"    current best cost = {cost}\n"

    data = np.array(data)
    matrix_output = data.T

    df_output = pd.DataFrame(matrix_output)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"
    if not save_inter:
        tsc.ul.cleanup(tmpdir)

    if not experiment:
        tsc.ul.stat(df_input, df_output, alpha, beta, running_time)
        # for k, v in detail.items():
        #     tsc.logg.info(f"{k}: {v}")
        return df_output
    else:
        return df_output, running_time


def rscistree(adata, mode="haploid", save_inter=False):
    tsc.logg.info(f"running rScisTree with mode={mode}")
    tmpdir = tsc.ul.tmpdir(prefix="trisicell.", suffix=".scistree", dirname=".")

    cells = adata.obs_names
    snvs = adata.var_names
    tsc.pp.build_scmatrix(adata)
    df_input = adata.to_df()

    V = adata.layers["mutant"]
    R = adata.layers["total"] - V
    with open(f"{tmpdir}/rscistree.counts", "w") as fout:
        fout.write(f"{mode.upper()} {len(snvs)} {len(cells)}\n")
        for j in range(len(snvs)):
            for i in range(len(cells)):
                fout.write(f"{R[i,j]} {V[i,j]}     ")
            fout.write(f"\n")
    cmd = f"/home/frashidi/software/temp/scistree/scprob/scprob_{mode.upper()} "
    cmd += f"{tmpdir}/rscistree.counts > {tmpdir}/rscistree.input"
    os.system(cmd)

    scistree = tsc.ul.get_file("trisicell.external/bin/scistree")
    cmd = (
        f"{scistree} "
        "-v "
        "-d 0 "
        "-e "
        f"-o {tmpdir}/rscistree.gml "
        f"{tmpdir}/rscistree.input > {tmpdir}/rscistree.output"
    )
    s_time = time.time()
    os.system(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    data = []
    mut_tree = ""
    cell_tree = ""
    detail = {"cost": "\n"}
    with open(f"{tmpdir}/rscistree.output") as infile:
        now_store = False
        for line in infile:
            line = line.strip()
            if "Mutation tree:" in line:
                mut_tree = line.split(":")[1].replace(" ", "")
                mut_tree = mut_tree.replace("#", "")
            if "Constructed single cell phylogeny:" in line:
                cell_tree = line.split(":")[1].replace(" ", "")
            if "Imputed genotypes:" in line:
                now_store = True
            if line[:4] == "Site" and now_store:
                line = "".join(line.split(":")[1])
                line = line.replace("\t", "")
                data.append([int(x) for x in line.split(" ")])
            if "current cost: " in line:
                cost = float(line.split("current cost: ")[1].split(", opt tree: ")[0])
                detail["cost"] += f"    current best cost = {cost}\n"

    data = np.array(data)
    matrix_output = data.T

    df_output = pd.DataFrame(matrix_output)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"
    if not save_inter:
        tsc.ul.cleanup(tmpdir)

    tsc.ul.stat(df_input, df_output, alpha, beta, running_time)
    # for k, v in detail.items():
    #     tsc.logg.info(f"{k}: {v}")

    return df_output


def iscistree(df_input, alpha, beta, n_iters=np.inf):
    tsc.logg.info(
        f"running iScisTree with alpha={alpha}, beta={beta}, n_iters={n_iters}"
    )

    def get_initial_tree(D):
        Q = []
        for i in range(D.shape[0]):
            Q.append(list(D[i, : i + 1]))
        constructor = DistanceTreeConstructor()
        dm = DistanceMatrix(names=[f"{i}" for i in range(D.shape[0])], matrix=Q)
        tree = nj(dm)
        # tree = constructor.nj(dm)
        # tree = constructor.upgma(dm)

        node = None
        for clade in tree.find_clades():
            if clade.name == f"{D.shape[0]-1}":
                node = clade
        tree.root_with_outgroup(node)
        tree.prune(f"{D.shape[0]-1}")
        return tree

    def get_subtrees(tree):
        subtrees = []
        for clade in tree.find_clades():
            gen = np.zeros(tree.count_terminals(), dtype=int)
            cel = [
                int(clade2.name)
                for clade2 in clade.find_clades()
                if "Inner" not in clade2.name
            ]
            gen[cel] = 1
            subtrees.append(gen)
        subtrees = np.array(subtrees)
        return subtrees

    def denoise_quadratic(I, alpha, beta, subtrees):
        def column_pairs_cost(A, Ap, unit_costs):
            num = np.zeros((2, 2), dtype=np.int)
            for i in range(2):
                for j in range(2):
                    num[i, j] = np.count_nonzero(np.logical_and(A == i, Ap == j))
            return np.sum(num * unit_costs)

        unit_prob = np.array([[1 - beta, beta], [alpha, 1 - alpha]])
        unit_costs = -np.log(unit_prob)
        output = np.zeros(I.shape, dtype=int)
        total_cost = 0
        for c in range(I.shape[1]):
            costs = [
                column_pairs_cost(I[:, c], subtrees[st_ind], unit_costs)
                for st_ind in range(len(subtrees))
            ]
            ind = np.argmin(costs)
            output[:, c] = subtrees[ind]
            total_cost += costs[ind]
        return output, total_cost

    def denoise_linear(I, alpha, beta, opt_tree):
        tree = {}
        for clade in list(opt_tree.find_clades(order="level"))[::-1]:
            children = list(clade.find_clades(order="level"))
            if len(children) > 2:
                child_l = children[2]
                child_r = children[1]
                tree[clade.name] = [child_l.name, child_r.name]
            else:
                tree[clade.name] = []

        def get_cells_in_best(cells_in_best, best):
            for node in tree[best]:
                if "Inner" in node:
                    for child in tree[node]:
                        get_cells_in_best(cells_in_best, child)
                else:
                    cells_in_best.append(node)
            if "Inner" not in best:
                cells_in_best.append(best)

        output = np.zeros(I.shape, dtype=int)
        total_cost = 0
        for c in range(I.shape[1]):
            qs = {}
            best = None
            best_v = 0
            for k, v in tree.items():
                if len(v) == 0:
                    obs = I[int(k), c] == 1
                    # qs[k] = (beta**(1-obs) + (1-beta)**obs) / (alpha**obs + (1-alpha)**(1-obs))
                    # qs[k] = np.log((beta ** (1 - obs)) / (alpha ** obs))
                    p0 = (1 - obs) * (1 - beta) + obs * alpha
                    qs[k] = np.log((1 - p0) / p0)
                else:
                    qs[k] = qs[v[0]] + qs[v[1]]
                if qs[k] > best_v:
                    best = k
                    best_v = qs[k]

            cells_in_best = []
            get_cells_in_best(cells_in_best, best)
            output[list(map(int, cells_in_best)), c] = 1

            # a = (I[:, c] == 0).sum() * np.log(1 - beta)
            # b = (I[:, c] == 1).sum() * np.log(alpha)
            # best_v += a + b
            total_cost += -best_v
        return output, total_cost

    def draw_helper(x):
        if x.is_terminal():
            return f"{int(x.name)+1}"
        else:
            return None

    def get_neighbors(tree):
        """
        For a tree with n taxa, there are n - 3 internal branches.
        Thus there are 2(n - 3) NNI rearrangements for any tree
        """
        # make child to parent dict
        parents = {}
        for clade in tree.find_clades():
            if clade != tree.root:
                node_path = tree.get_path(clade)
                # cannot get the parent if the parent is root. Bug?
                if len(node_path) == 1:
                    parents[clade] = tree.root
                else:
                    parents[clade] = node_path[-2]
        neighbors = []
        root_childs = []
        for clade in tree.get_nonterminals(order="level"):
            if clade == tree.root:
                left = clade.clades[0]
                right = clade.clades[1]
                root_childs.append(left)
                root_childs.append(right)
                if not left.is_terminal() and not right.is_terminal():
                    # make changes around the left_left clade
                    # left_left = left.clades[0]
                    left_right = left.clades[1]
                    right_left = right.clades[0]
                    right_right = right.clades[1]
                    # neightbor 1 (left_left + right_right)
                    del left.clades[1]
                    del right.clades[1]
                    left.clades.append(right_right)
                    right.clades.append(left_right)
                    temp_tree = copy.deepcopy(tree)
                    neighbors.append(temp_tree)
                    # neighbor 2 (left_left + right_left)
                    del left.clades[1]
                    del right.clades[0]
                    left.clades.append(right_left)
                    right.clades.append(right_right)
                    temp_tree = copy.deepcopy(tree)
                    neighbors.append(temp_tree)
                    # change back (left_left + left_right)
                    del left.clades[1]
                    del right.clades[0]
                    left.clades.append(left_right)
                    right.clades.insert(0, right_left)
            elif clade in root_childs:
                # skip root child
                continue
            else:
                # method for other clades
                # make changes around the parent clade
                left = clade.clades[0]
                right = clade.clades[1]
                parent = parents[clade]
                if clade == parent.clades[0]:
                    sister = parent.clades[1]
                    # neighbor 1 (parent + right)
                    del parent.clades[1]
                    del clade.clades[1]
                    parent.clades.append(right)
                    clade.clades.append(sister)
                    temp_tree = copy.deepcopy(tree)
                    neighbors.append(temp_tree)
                    # neighbor 2 (parent + left)
                    del parent.clades[1]
                    del clade.clades[0]
                    parent.clades.append(left)
                    clade.clades.append(right)
                    temp_tree = copy.deepcopy(tree)
                    neighbors.append(temp_tree)
                    # change back (parent + sister)
                    del parent.clades[1]
                    del clade.clades[0]
                    parent.clades.append(sister)
                    clade.clades.insert(0, left)
                else:
                    sister = parent.clades[0]
                    # neighbor 1 (parent + right)
                    del parent.clades[0]
                    del clade.clades[1]
                    parent.clades.insert(0, right)
                    clade.clades.append(sister)
                    temp_tree = copy.deepcopy(tree)
                    neighbors.append(temp_tree)
                    # neighbor 2 (parent + left)
                    del parent.clades[0]
                    del clade.clades[0]
                    parent.clades.insert(0, left)
                    clade.clades.append(right)
                    temp_tree = copy.deepcopy(tree)
                    neighbors.append(temp_tree)
                    # change back (parent + sister)
                    del parent.clades[0]
                    del clade.clades[0]
                    parent.clades.insert(0, sister)
                    clade.clades.insert(0, left)
        return neighbors

    def nj(distance_matrix):
        # make a copy of the distance matrix to be used
        dm = copy.deepcopy(distance_matrix)
        # init terminal clades
        clades = [BaseTree.Clade(None, name) for name in dm.names]
        # init node distance
        node_dist = [0] * len(dm)
        # init minimum index
        min_i = 0
        min_j = 0
        inner_count = 0
        # special cases for Minimum Alignment Matrices
        if len(dm) == 1:
            root = clades[0]
            return BaseTree.Tree(root, rooted=False)
        elif len(dm) == 2:
            # minimum distance will always be [1,0]
            min_i = 1
            min_j = 0
            clade1 = clades[min_i]
            clade2 = clades[min_j]
            clade1.branch_length = dm[min_i, min_j] / 2.0
            clade2.branch_length = dm[min_i, min_j] - clade1.branch_length
            inner_clade = BaseTree.Clade(None, "Inner")
            inner_clade.clades.append(clade1)
            inner_clade.clades.append(clade2)
            clades[0] = inner_clade
            root = clades[0]
            return BaseTree.Tree(root, rooted=False)
        while len(dm) > 2:
            # calculate nodeDist
            for i in range(0, len(dm)):
                node_dist[i] = 0
                for j in range(0, len(dm)):
                    node_dist[i] += dm[i, j]
                node_dist[i] = node_dist[i] / (len(dm) - 2)

            # find minimum distance pair
            min_dist = dm[1, 0] - node_dist[1] - node_dist[0]
            min_i = 0
            min_j = 1
            for i in range(1, len(dm)):
                for j in range(0, i):
                    temp = dm[i, j] - node_dist[i] - node_dist[j]
                    if min_dist > temp:
                        min_dist = temp
                        min_i = i
                        min_j = j
            # create clade
            clade1 = clades[min_i]
            clade2 = clades[min_j]
            inner_count += 1
            inner_clade = BaseTree.Clade(None, "Inner" + str(inner_count))
            inner_clade.clades.append(clade1)
            inner_clade.clades.append(clade2)
            # assign branch length
            clade1.branch_length = (
                dm[min_i, min_j] + node_dist[min_i] - node_dist[min_j]
            ) / 2.0
            clade2.branch_length = dm[min_i, min_j] - clade1.branch_length
            if clade1.branch_length < 0:
                clade1.branch_length = 0
            if clade2.branch_length < 0:
                clade2.branch_length = 0

            # update node list
            clades[min_j] = inner_clade
            del clades[min_i]

            # rebuild distance matrix,
            # set the distances of new node at the index of min_j
            for k in range(0, len(dm)):
                if k != min_i and k != min_j:
                    dm[min_j, k] = (
                        dm[min_i, k] + dm[min_j, k] - dm[min_i, min_j]
                    ) / 2.0

            dm.names[min_j] = "Inner" + str(inner_count)
            del dm[min_i]

        # set the last clade as one of the child of the inner_clade
        root = None
        if clades[0] == inner_clade:
            clades[0].branch_length = 0
            clades[1].branch_length = dm[1, 0]
            clades[0].clades.append(clades[1])
            root = clades[0]
        else:
            clades[0].branch_length = dm[1, 0]
            clades[1].branch_length = 0
            clades[1].clades.append(clades[0])
            root = clades[1]

        return BaseTree.Tree(root, rooted=False)

    cells = list(df_input.index)
    snvs = list(df_input.columns)
    I = df_input.values

    s_time = time.time()
    Ip = np.vstack([I, np.zeros(I.shape[1])])  # add root with profile zero
    dist = tsc.ul.calc_dist_by_l1_ignore_na(Ip)
    opt_tree = get_initial_tree(dist)
    # opt_subtrees = get_subtrees(opt_tree)
    # opt_O, opt_cost = denoise_quadratic(I, alpha, beta, opt_subtrees)
    opt_O, opt_cost = denoise_linear(I, alpha, beta, opt_tree)
    tsc.logg.info("current best cost =", opt_cost, time=True)

    n_iter = 1
    is_done = False
    already_seen = set()
    already_seen.add(str(opt_tree))
    while not is_done and n_iter < n_iters:
        is_done = True
        neighbors = get_neighbors(opt_tree)
        for nbr_tree in neighbors:
            if str(nbr_tree) in already_seen:
                continue
            else:
                already_seen.add(str(nbr_tree))
            # nbr_subtrees = get_subtrees(nbr_tree)
            # nbr_O, nbr_cost = denoise_quadratic(I, alpha, beta, nbr_subtrees)
            nbr_O, nbr_cost = denoise_linear(I, alpha, beta, nbr_tree)
            if nbr_cost < opt_cost:
                opt_tree = nbr_tree
                # opt_subtrees = nbr_subtrees
                opt_O = nbr_O
                opt_cost = nbr_cost
                is_done = False
                tsc.logg.info("current best cost =", nbr_cost, time=True)
        n_iter += 1
    e_time = time.time()
    running_time = e_time - s_time

    df_output = pd.DataFrame(opt_O)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"

    tsc.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



import datetime
import os
import shutil
import tempfile
import time

import numpy as np
import pandas as pd
import pkg_resources

import trisicell as tsc


def log_input(df_in):
    size = df_in.shape[0] * df_in.shape[1]
    tsc.logg.info(f"input -- size: {df_in.shape[0]}x{df_in.shape[1]}")
    tsc.logg.info(
        f"input -- 0: {np.sum(df_in.values == 0)}#,"
        f" {100*np.sum(df_in.values == 0)/size:.1f}%"
    )
    tsc.logg.info(
        f"input -- 1: {np.sum(df_in.values == 1)}#,"
        f" {100*np.sum(df_in.values == 1)/size:.1f}%"
    )
    tsc.logg.info(
        f"input -- NA: {np.sum(df_in.values == 3)}#,"
        f" {100*np.sum(df_in.values == 3)/size:.1f}%"
    )
    tsc.logg.info(f"input -- CF: {is_conflict_free_gusfield(df_in)}")


def log_output(df_out, running_time):
    size = df_out.shape[0] * df_out.shape[1]
    tsc.logg.info(f"output -- size: {df_out.shape[0]}x{df_out.shape[1]}")
    tsc.logg.info(
        f"output -- 0: {np.sum(df_out.values == 0)}#,"
        f" {100*np.sum(df_out.values == 0)/size:.1f}%"
    )
    tsc.logg.info(
        f"output -- 1: {np.sum(df_out.values == 1)}#,"
        f" {100*np.sum(df_out.values == 1)/size:.1f}%"
    )
    tsc.logg.info(
        f"output -- NA: {np.sum(df_out.values == 3)}#,"
        f" {100*np.sum(df_out.values == 3)/size:.1f}%"
    )
    tsc.logg.info(f"output -- CF: {is_conflict_free_gusfield(df_out)}")
    tsc.logg.info(
        f"output -- time: {running_time:.1f}s"
        f" ({datetime.timedelta(seconds=running_time)})"
    )


def log_flip(df_in, df_out):
    flips_0_1, flips_1_0, flips_na_0, flips_na_1 = count_flips(
        df_in.values, df_out.values, 3
    )
    fn_rate, fp_rate, na_rate = infer_rates(df_in.values, df_out.values, 3)
    tsc.logg.info(f"flips -- #0->1: {flips_0_1}")
    tsc.logg.info(f"flips -- #1->0: {flips_1_0}")
    tsc.logg.info(f"flips -- #NA->0: {flips_na_0}")
    tsc.logg.info(f"flips -- #NA->1: {flips_na_1}")
    tsc.logg.info(f"rates -- FN: {fn_rate:.3f}")
    tsc.logg.info(f"rates -- FP: {fp_rate:.8f}")
    tsc.logg.info(f"rates -- NA: {na_rate:.3f}")


def calc_score_tree(df_in, df_out, alpha, beta):
    if alpha == 0 or beta == 0:
        return None
    columns = np.intersect1d(df_in.columns, df_out.columns)
    indices = np.intersect1d(df_in.index, df_out.index)
    D = df_in.loc[indices, columns].values
    E = df_out.loc[indices, columns].values
    removedMutations = []
    objective = 0
    for j in range(len(columns)):
        numZeros = 0
        numOnes = 0
        for i in range(len(indices)):
            if D[i, j] == 0:
                numZeros += 1
                objective += np.log(beta / (1 - alpha)) * E[i, j]
            elif D[i, j] == 1:
                numOnes += 1
                objective += np.log((1 - beta) / alpha) * E[i, j]
        objective += numZeros * np.log(1 - alpha)
        objective += numOnes * np.log(alpha)
        if j in removedMutations:
            objective -= numZeros * np.log(1 - alpha) + numOnes * (
                np.log(alpha) + np.log((1 - beta) / alpha)
            )
    tsc.logg.info(f"score -- NLL: {-objective}")
    return None


def stat(df_in, df_out, alpha, beta, running_time):
    log_input(df_in)
    log_output(df_out, running_time)
    log_flip(df_in, df_out)
    calc_score_tree(df_in, df_out, alpha, beta)


def get_param(filename):
    data = {}
    # simNo_2-s_7-m_20-h_1-minVAF_0.1-ISAV_0-n_10-fp_0-fn_0.1-na_0-d_0-l_1000000.SC
    dirname, basename = dir_base(filename)
    data["simNo"] = int(basename.split("-")[0].split("_")[1])
    data["s"] = int(basename.split("-")[1].split("_")[1])
    data["m"] = int(basename.split("-")[2].split("_")[1])
    data["h"] = int(basename.split("-")[3].split("_")[1])
    data["minVAF"] = float(basename.split("-")[4].split("_")[1])
    data["ISAV"] = int(basename.split("-")[5].split("_")[1])
    data["n"] = int(basename.split("-")[6].split("_")[1])
    data["fp"] = float(basename.split("-")[7].split("_")[1])
    data["fn"] = float(basename.split("-")[8].split("_")[1])
    data["na"] = float(basename.split("-")[9].split("_")[1])
    data["d"] = int(basename.split("-")[10].split("_")[1])
    last = basename.split("-")[11]
    if "." in last:
        data["l"] = int(last.split(".")[0].split("_")[1])
    else:
        data["l"] = int(last.split("_")[1])
    return data


def count_flips(I, O, na_value=3):
    flips_0_1 = 0
    flips_1_0 = 0
    flips_na_0 = 0
    flips_na_1 = 0
    n, m = I.shape
    for i in range(n):
        for j in range(m):
            if I[i, j] == 0 and O[i, j] == 1:
                flips_0_1 += 1
            elif I[i, j] == 1 and O[i, j] == 0:
                flips_1_0 += 1
            elif I[i, j] == na_value and O[i, j] == 0:
                flips_na_0 += 1
            elif I[i, j] == na_value and O[i, j] == 1:
                flips_na_1 += 1
    return flips_0_1, flips_1_0, flips_na_0, flips_na_1


def infer_rates(I, O, na_value=3):
    flips_0_1, flips_1_0, flips_na_0, flips_na_1 = count_flips(I, O, na_value)
    fn_rate = flips_0_1 / ((O == 1) & (I != na_value)).sum()
    fp_rate = flips_1_0 / ((O == 0) & (I != na_value)).sum()
    na_rate = (flips_na_1 + flips_na_0) / I.size
    return fn_rate, fp_rate, na_rate


def is_conflict_free(D, na_value=3):
    conflict_free = True
    for p in range(D.shape[1]):
        for q in range(p + 1, D.shape[1]):
            oneone = False
            zeroone = False
            onezero = False
            for r in range(D.shape[0]):
                if D[r, p] == 1 and D[r, q] == 1:
                    oneone = True
                if D[r, p] == 0 and D[r, q] == 1:
                    zeroone = True
                if D[r, p] == 1 and D[r, q] == 0:
                    onezero = True
            if oneone and zeroone and onezero:
                conflict_free = False
                return conflict_free
    return conflict_free


def is_conflict_free_gusfield(df_in, na_value=3):
    """Check conflict-free criteria via Gusfield algorithm.

    This is an implementation of algorithm 1.1 in :cite:`Gusfield_1991`.

    The order of this algorithm is O(nm)
    where n is the number of cells and m is the number of mutations.

    Parameters
    ----------
    df_in : :class:`pandas.DataFrame`
        Input genotype matrix.
    na_value : :obj:`int`, optional
        Missing value representation in the input data, by default 3.

    Returns
    -------
    :obj:`bool`
        A Boolean checking if the input conflict-free or not.

    Examples
    --------
    >>> sc = tsc.datasets.test()
    >>> tsc.ul.is_conflict_free_gusfield(sc)
    False

    See Also
    --------
    :func:`trisicell.ul.is_conflict_free`.
    """

    I = df_in.astype(int).values

    def sort_bin(a):
        b = np.transpose(a)
        b_view = np.ascontiguousarray(b).view(
            np.dtype((np.void, b.dtype.itemsize * b.shape[1]))
        )
        idx = np.argsort(b_view.ravel())[::-1]
        c = b[idx]
        return np.transpose(c), idx

    Ip = I.copy()
    # Ip[Ip == na_value] = 0
    O, idx = sort_bin(Ip)
    # tsc.logg.info(O, '\n')
    Lij = np.zeros(O.shape, dtype=int)
    for i in range(O.shape[0]):
        maxK = 0
        for j in range(O.shape[1]):
            if O[i, j] == 1:
                Lij[i, j] = maxK
                maxK = j + 1
    # tsc.logg.info(Lij, '\n')
    Lj = np.amax(Lij, axis=0)
    # tsc.logg.info(Lj, '\n')
    for i in range(O.shape[0]):
        for j in range(O.shape[1]):
            if O[i, j] == 1:
                if Lij[i, j] != Lj[j]:
                    return False  # , (idx[j], idx[Lj[j] - 1])
    return True  # , (None, None)


def tmpdir(prefix="trisicell_", suffix="_trisicell", dirname="."):
    return tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dirname)


def tmpfile(prefix="trisicell_", suffix="_trisicell", dirname="."):
    return tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=dirname)[1]


def cleanup(dirname):
    shutil.rmtree(dirname)


def remove(filename):
    if os.path.exists(filename):
        os.remove(filename)


def dir_base(infile):
    basename = os.path.splitext(os.path.basename(infile))[0]
    dirname = os.path.dirname(infile)
    return dirname, basename


def dirbase(infile):
    return os.path.splitext(infile)[0]


def mkdir(indir):
    if not os.path.exists(indir):
        os.makedirs(indir)
    return indir


def timeit(f):
    def wrap(*args, **kwargs):
        start_time = time.time()
        ret = f(*args, **kwargs)
        end_time = time.time()
        if end_time - start_time < 60:
            tsc.logg.info(f"Time needed for {f.__name__}: {end_time - start_time:.3f}")
        else:
            tsc.logg.info(
                f"Time needed for {f.__name__}:"
                f" {time.strftime('%Hh:%Mm:%Ss', time.gmtime(end_time - start_time))}"
            )
        return ret

    return wrap


def get_file(key):
    components = key.split("/")
    return pkg_resources.resource_filename(components[0], "/".join(components[1:]))


def split_mut(mut):
    alt = mut.split(".")[-1]
    try:
        ref = mut.split(".")[-2]
    except:
        return None, None, None, None, None, None
    pos = mut.split(".")[-3]
    chrom = mut.split(".")[-4]
    gene = None
    ens = None
    try:
        gene = mut.split(".chr")[0].split("_")[1]
        ens = mut.split(".chr")[0].split("_")[0]
    except:
        pass
    return ens, gene, chrom, pos, ref, alt


def flips_in_cells(adata, df_in, df_out):
    c01 = ((df_in == 0) & (df_out == 1)).sum(axis=1)
    c10 = ((df_in == 1) & (df_out == 0)).sum(axis=1)
    c31 = ((df_in == 3) & (df_out == 1)).sum(axis=1)
    c30 = ((df_in == 3) & (df_out == 0)).sum(axis=1)
    c01 = c01 / c01.sum()
    c10 = c10 / c10.sum()
    c31 = c31 / c31.sum()
    c30 = c30 / c30.sum()
    c01 = c01.rename("f_0_1")
    c10 = c10.rename("f_1_0")
    c31 = c31.rename("f_3_1")
    c30 = c30.rename("f_3_0")
    df = pd.DataFrame([c01, c10, c31, c30]).T
    df["f_0_1_color"] = "#E41A1C"
    df["f_1_0_color"] = "#FF7F00"
    df["f_3_1_color"] = "#FDBF6F"
    df["f_3_0_color"] = "#FB9A99"
    adata.obs = pd.merge(adata.obs, df, how="left", left_index=True, right_index=True)


def get_database(species):
    data = []
    temp = {}
    with open(f"/home/frashidi/database/{species}/ensembl/{species}.rsem") as fin:
        i = -2
        for line in fin:
            i += 1
            line = line.strip()
            if i == -1:
                continue
            if i % 6 == 0:
                temp = {}
                temp["trans_ens"] = line.split("\t")[0]
                temp["trans_sym"] = line.split("\t")[1]
            if i % 6 == 1:
                temp["gene_ens"] = line.split("\t")[0]
                temp["gene_sym"] = line.split("\t")[1]
            if i % 6 == 2:
                temp["chrom"] = line
            if i % 6 == 3:
                temp["strand"] = line.split(" ")[0]
            if i % 6 == 4:
                temp["n_exons"] = int(line.split(" ")[0])
                temp["regions"] = line.split(" ")[1:]
                temp["start"] = int(line.split(" ")[1])
                temp["end"] = int(line.split(" ")[-1])
            if i % 6 == 5:
                details = line.split("; ")
                for det in details:
                    temp[det.split(" ")[0]] = det.split(" ")[1]
                data.append(temp)
    database = pd.DataFrame(data)
    return database


def translate_gene(alist, kind):
    mm10 = get_database("mm10")
    mm10 = mm10.set_index("gene_sym")["gene_ens"].drop_duplicates().to_dict()
    hg19 = get_database("hg19")
    hg19 = hg19.set_index("gene_sym")["gene_ens"].drop_duplicates().to_dict()
    if kind == "mm10":
        translator = mm10
    else:
        translator = hg19
    result = []
    for d in alist:
        if d in translator:
            result.append(f"{translator[d]}_{d}")
        else:
            tsc.logg.info(d)
    return result


def difference_between(dna, rna):
    dna_temp = (
        dna.var["chrom"].astype(str)
        + ":"
        + dna.var["position"].astype(str)
        + ":"
        + dna.var["reference"].astype(str)
        + ":"
        + dna.var["alteration"].astype(str)
    )
    rna_temp = (
        rna.var["chrom"].astype(str)
        + ":"
        + rna.var["position"].astype(str)
        + ":"
        + rna.var["reference"].astype(str)
        + ":"
        + rna.var["alteration"].astype(str)
    )
    a = np.intersect1d(dna_temp.values, rna_temp.values)
    b = np.setdiff1d(dna_temp.values, rna_temp.values)
    c = np.setdiff1d(rna_temp.values, dna_temp.values)
    tsc.logg.info(f"{len(b)}, inter={len(a)}, {len(c)}")

    b = dna.to_df(layer="mutant").T.loc[dna_temp[dna_temp.isin(b)].index]
    c = rna.to_df(layer="mutant").T.loc[rna_temp[rna_temp.isin(c)].index]

    return dna_temp[dna_temp.isin(a)].index, b, c



import glob
import operator
import os
import pickle
import sys

import anndata as ad
import numpy as np
import pandas as pd
from cyvcf2 import VCF

import trisicell as tsc


def read(filepath):
    """Read genotype matrix and read-count matrix.

    The genotype matrix must be in the in format of :class:`pandas.DataFrame`
    The read-count matrix must be in the format of :class:`anndata.AnnData`.

    Parameters
    ----------
    filepath : :obj:`str`
        The path to the file. The extension must be one of
        [`.SC`, `.CFMatrix`, `.h5ad`, `.h5ad.gz`]

    Returns
    -------
    :class:`pandas.DataFrame` or :class:`anndata.AnnData`
        Depends on the format of the input file the output type is different.

    Raises
    ------
    ValueError
        If the extension is not one of the above.
    """

    ext = os.path.splitext(filepath)[-1]
    if ext in [".SC", ".CFMatrix", ".before_FP_FN_NA"]:
        sc = pd.read_table(filepath, index_col=0)
        if len(sc.columns) != len(set(sc.columns)):
            raise RuntimeError("Mutation ids must be unique!")
        return sc
    elif ext in [".h5ad", ".gz"]:
        return ad.read(filepath)
    else:
        raise ValueError("Extension is wrong!")


def write(obj, filepath):
    """Write genotype matrix or read-count matrix into a file.

    Parameters
    ----------
    obj : :class:`pandas.DataFrame` or :class:`anndata.AnnData`
        The input object which is going to be written in a file.
    filepath : :obj:`str`
        The file path where the `obj` must be written in.

    Raises
    ------
    ValueError
        If `obj` is not an instance of:
            :class:`pandas.DataFrame`
            :class:`anndata.AnnData`
    """

    if isinstance(obj, pd.DataFrame):
        obj.index.name = "cellIDxmutID"
        obj.to_csv(filepath, sep="\t")
    elif isinstance(obj, ad.AnnData):
        obj.write(filepath + ".h5ad.gz", compression="gzip")
    else:
        raise ValueError("Object instance is wrong!")


def read_gatk(
    folderpath,
    genome_id,
    only_snp=False,
    only_exonic=True,
    filter_private=False,
    filter_dbsnp=True,
    filter_indels_greater_than_2=True,
):
    """Read GATK genotype output to AnnData matrix.

    Parameters
    ----------
    folderpath : :obj:`str`
        The path to the folder contains following files:

        1) `jointcalls.g.vcf`
        2) `exonic.tsv`
        3) `readstat.tsv`
    genome_id : :obj:`str`
        One of `hg19`, `mm10`
    only_snp : :obj:`bool`, optional
        If only SNPs must be called (i.e. Indels are ignored), by default False
    only_exonic : :obj:`bool`, optional
        Only exonic mutations are considered, by default True
    filter_private : :obj:`bool`, optional
        Filter private mutation (i.e. only present in one cell), by default False
    filter_dbsnp : :obj:`bool`, optional
        Filter mutations that are present in dbSNP, by default True
    filter_indels_greater_than_2 : :obj:`bool`, optional
        Filter Indels that the length is greater than 2, by default True

    Returns
    -------
    :class:`anndata.AnnData`
        The AnnData object which includes layers of mutant, total and genotype
    """

    def get_germline(species):
        filename = f"/home/frashidi/database/{species}/dbsnp/{species}.pickle"
        obj = pickle.load(open(filename, "rb"))
        return obj

    def get_database(species):
        data = []
        with open(f"/home/frashidi/database/{species}/ensembl/{species}.rsem") as fin:
            i = -2
            for line in fin:
                i += 1
                line = line.strip()
                if i == -1:
                    continue
                if i % 6 == 1:
                    symbol = line.split("\t")[1]
                if i % 6 == 2:
                    chrom = line
                if i % 6 == 3:
                    strand = line.split(" ")[0]
                if i % 6 == 5:
                    details = line.split(";")
                    ensg = details[0].split("gene_id ")[1].replace('"', "")
                    enst = details[1].split("transcript_id ")[1].replace('"', "")
                    data.append(
                        {
                            "enst_symbol": f"{enst}_{symbol}",
                            "ensg_symbol": f"{ensg}_{symbol}",
                        }
                    )
        database = pd.DataFrame(data).set_index("enst_symbol")["ensg_symbol"].to_dict()
        return database

    def get_exonic(exonic_file, species):
        database = get_database(species)
        df = pd.read_table(exonic_file, header=None)
        df["muts"] = df.apply(lambda x: f"{x[1]}.{x[2]}.{x[3]}.{x[4]}", axis=1)
        df["gene"] = df[0].map(database)
        df["kind"] = df[5]
        df["amino_acid_change"] = df[6]
        df = df.drop([0, 1, 2, 3, 4, 5, 6], axis=1)
        df = df.drop_duplicates("muts")
        return df.set_index("muts")

    vcf_file = f"{folderpath}/jointcalls.g.vcf"
    exonic_file = f"{folderpath}/exonic.tsv"
    if genome_id == "human":
        species = "hg19"
        germline = get_germline(species)
    else:
        species = "mm10"
        germline = get_germline(species)

    if not filter_dbsnp:
        germline = []

    databse = get_database(species)
    exonic = get_exonic(exonic_file, species)
    translator = exonic["gene"].to_dict()

    # vcf = VCF(vcf_file, samples=list(info.index))
    vcf = VCF(vcf_file)
    cells = vcf.samples
    cells = pd.Series(cells, index=range(len(cells)), dtype=str, name="cells")
    n_total = 0
    n_multiallelic = 0
    n_germline = 0
    n_exonic = 0
    filtered = []
    for v in vcf():
        n_total += 1
        index = ""
        if len(v.ALT) != 1:
            n_multiallelic += 1
            continue
        if v.is_snp:
            index = f"{v.CHROM}.{v.start+1}.{v.REF}.{v.ALT[0]}"
        if not only_snp:
            if v.is_indel:
                a = v.REF
                b = v.ALT[0]
                if v.var_subtype == "del":
                    a = a[len(b) :]
                    b = "-"
                    index = f"{v.CHROM}.{v.start+2}.{a}.{b}"
                elif v.var_subtype == "ins":
                    a = "-"
                    b = b[len(a) :]
                    index = f"{v.CHROM}.{v.start+1}.{a}.{b}"
                if filter_indels_greater_than_2:
                    if len(a) > 2:
                        continue
                    if len(b) > 2:
                        continue
        if index in germline:
            n_germline += 1
            continue
        if only_exonic:
            if index not in translator:
                n_exonic += 1
                continue
        if filter_private:
            if sum(v.gt_types == 1) + sum(v.gt_types == 3) < 2:
                continue
        filtered.append(v)

    tsc.logg.info(f"{n_total} variants reported by GATK")
    tsc.logg.info(f"filtered {n_multiallelic} due to being multi allelic")
    tsc.logg.info(f"filtered {n_germline} due to being germline")
    tsc.logg.info(f"filtered {n_exonic} due to not being exonic")

    muts = []
    for v in filtered:
        if v.is_snp:
            index = f"{v.CHROM}.{v.start+1}.{v.REF}.{v.ALT[0]}"
        if v.is_indel:
            a = v.REF
            b = v.ALT[0]
            if v.var_subtype == "del":
                a = a[len(b) :]
                b = "-"
                index = f"{v.CHROM}.{v.start+2}.{a}.{b}"
            elif v.var_subtype == "ins":
                a = "-"
                b = b[len(a) :]
                index = f"{v.CHROM}.{v.start+1}.{a}.{b}"
        if only_exonic:
            mut = f"{translator[index]}.{index}"
        else:
            mut = index
        muts.append(mut)
    muts = pd.Series(muts, index=range(len(muts)), dtype=str, name="muts")

    if only_exonic:
        exonic = exonic.reset_index()
        exonic["muts"] = exonic["gene"] + "." + exonic["muts"]
        exonic = exonic[["muts", "kind", "amino_acid_change"]]
        exonic[["ensemble", "gene", "chrom", "position", "reference", "alteration"]] = (
            exonic["muts"].apply(tsc.ul.split_mut).apply(pd.Series)
        )
        exonic = exonic.set_index("muts")

    X = 3 * np.ones((len(cells), len(muts)), dtype=int)
    T = np.zeros((len(cells), len(muts)), dtype=int)
    V = np.zeros((len(cells), len(muts)), dtype=int)
    G = np.zeros((len(cells), len(muts)), dtype=int)

    for j, v in enumerate(filtered):
        T[:, j] = v.gt_ref_depths + v.gt_alt_depths
        V[:, j] = v.gt_alt_depths
        G[:, j] = v.gt_types

    adata = ad.AnnData(
        X=X,
        obs=pd.DataFrame(cells).set_index("cells"),
        var=exonic.loc[muts],
        layers={"total": T, "mutant": V, "genotype": G},
    )
    if os.path.exists(f"{folderpath}/readstat.tsv"):
        df = pd.read_table(f"{folderpath}/readstat.tsv", index_col=0)
        adata.obs = pd.merge(
            adata.obs, df, how="left", right_index=True, left_index=True
        )

    tsc.logg.info(adata)
    return adata


def read_rsem(folderpath):
    """Read RSEM expression output to AnnData matrix.

    Parameters
    ----------
    folderpath : :obj:`str`
        Path to the folder contains the following files:

        1) `expr.count.tsv`
        2) `expr.fpkm.tsv`
        3) `expr.tpm.tsv`
        4) `readstat.tsv`

    Returns
    -------
    :class:`anndata.AnnData`
        The AnnData object which includes layers of FPKM and TPM values
        and count in .X values.
    """

    count = pd.read_table(f"{folderpath}/expr.count.tsv", index_col=0)
    tpm = pd.read_table(f"{folderpath}/expr.tpm.tsv", index_col=0)
    fpkm = pd.read_table(f"{folderpath}/expr.fpkm.tsv", index_col=0)
    info = pd.read_table(f"{folderpath}/readstat.tsv", index_col=0)
    info.index.name = "cells"

    adata = ad.AnnData(
        X=count.loc[info.index].values,
        obs=info,
        var=pd.DataFrame.from_dict({"genes": count.columns}).set_index("genes"),
        layers={"tpm": tpm.loc[info.index], "fpkm": fpkm.loc[info.index]},
    )

    tsc.logg.info(adata)
    return adata


def read_cnvkit(folderpath):
    data = {}
    for file in glob.glob(f"{folderpath}/*.cnr"):
        name = file.split("/")[-1].replace(".cnr", "")
        df = pd.read_table(file)
        df = df[(df.gene != "Antitarget")]
        df = df.dropna()
        data[name] = df

    n = len(data.keys())
    m = data[list(data.keys())[0]].shape[0]
    depth = np.zeros((n, m))
    log2 = np.zeros((n, m))
    weight = np.zeros((n, m))
    X = np.zeros((n, m))

    for i, cell in enumerate(data.keys()):
        depth[i, :] = data[cell].depth.values
        log2[i, :] = data[cell].log2.values
        weight[i, :] = data[cell].weight.values

    depth[-1, :] = np.zeros(m)
    log2[-1, :] = np.zeros(m)
    weight[-1, :] = np.zeros(m)

    adata = ad.AnnData(
        X=X,
        obs=pd.DataFrame(index=list(data.keys())),
        var=data[list(data.keys())[0]][["chromosome", "start", "end", "gene"]],
        layers={"depth": depth, "log2": log2, "weight": weight},
    )

    cnv = np.zeros(adata.shape, dtype=int)
    tmp = adata.layers["log2"] < -1.1
    cnv[tmp == True] = 0
    tmp = (-1.1 <= adata.layers["log2"]) & (adata.layers["log2"] < -0.25)
    cnv[tmp == True] = 1
    tmp = (-0.25 <= adata.layers["log2"]) & (adata.layers["log2"] < 0.2)
    cnv[tmp == True] = 2
    tmp = (0.2 <= adata.layers["log2"]) & (adata.layers["log2"] < 0.7)
    cnv[tmp == True] = 3
    tmp = 0.7 <= adata.layers["log2"]
    cnv[tmp == True] = 4
    adata.layers["cnv"] = cnv

    abr = np.zeros(adata.shape, dtype=int)
    tmp = cnv < 2
    abr[tmp == True] = -1
    tmp = cnv == 2
    abr[tmp == True] = 0
    tmp = 2 < cnv
    abr[tmp == True] = 1
    adata.layers["abr"] = abr

    return adata


def read_bamreadcount(filepath):
    def get_nucleotides(line):
        chrom = line.split()[0]
        pos = line.split()[1]
        ref = line.split()[2].upper()
        depth = int(line.split()[3])
        char = {"A": 0, "C": 0, "G": 0, "T": 0, "N": 0}
        char[ref] = int(line.split()[4].split(":")[1])
        char[line.split()[5].split(":")[0]] = int(line.split()[5].split(":")[1])
        char[line.split()[6].split(":")[0]] = int(line.split()[6].split(":")[1])
        char[line.split()[7].split(":")[0]] = int(line.split()[7].split(":")[1])
        char[line.split()[8].split(":")[0]] = int(line.split()[8].split(":")[1])
        nucleotides = " ".join([str(char[a]) for a in ["A", "C", "G", "T", "N"]])
        cov = sum([char[a] for a in ["A", "C", "G", "T", "N"]])
        sorted_char = sorted(char.items(), key=operator.itemgetter(1), reverse=True)
        return ref, char, cov

    data = []
    for file in glob.glob(f"{filepath}/*.txt"):
        with open(file) as fin:
            name = file.split("/")[-1].replace(".txt", "")
            for line in fin:
                line = line.strip()
                a = ".".join(line.split("\t")[:3])
                ref, char, cov = get_nucleotides(line)
                res = {"name": name, "mut": a, "ref": ref, "cov": cov}
                for k, v in char.items():
                    res[k] = v
                data.append(res)
    df = pd.DataFrame(data)
    df["index"] = df.apply(lambda x: f"{x['name']}.{x['mut']}", axis=1)

    return df


def read_sc_bulk_simulation(sc_file, bulk_file):
    """Read single-cell and bulk daat from Salem's simulation
    and merge them as an AnnData object.

    Parameters
    ----------
    sc_file : :obj:`str`
        Path to single-cell file.
    bulk_file : :obj:`str`
        Path to bulk file.

    Returns
    -------
    :class:`anndata.AnnData`
        An object contains VAF information in `.var`.
    """

    sc = tsc.io.read(sc_file)
    bulk = pd.read_table(bulk_file)
    bulk = bulk.set_index("ID")
    bulk[["trueNode", "trueVAF", "ID", "sampleIDs"]] = (
        bulk.INFO.str.split(";")
        .apply(pd.Series)[[0, 1, 2, 3]]
        .rename(columns={0: "trueNode", 1: "trueVAF", 2: "ID", 3: "sampleIDs"})
    )
    bulk = bulk.drop(["INFO"], axis=1)
    bulk.trueNode = bulk.trueNode.str.replace("trueNode=", "")
    bulk.trueVAF = bulk.trueVAF.str.replace("trueVAF=", "")
    bulk.ID = bulk.ID.str.replace("ID=", "")
    bulk.sampleIDs = bulk.sampleIDs.str.replace("sampleIDs=", "")
    adata = ad.AnnData(sc, dtype=int)
    adata.var = bulk

    return adata


def read_snpeff(filepath):
    """Read the VCF file annotated by SnpEff in multi-sample format.

    SnpEff was introduced in :cite:`SnpEff`.

    Parameters
    ----------
    filepath : :obj:`str`
        The path to the VCF file.

    Returns
    -------
    :class:`anndata.AnnData`
        The AnnData object which includes layers of mutant, total and genotype
    """

    vcf = VCF(filepath)
    info = vcf.get_header_type("ANN")["Description"].split(" | ")
    info[0] = "Allele"
    info[-1] = "ERRORS / WARNINGS / INFO"
    info = ["CHROM", "POS", "REF", "ALT"] + info

    cells = vcf.samples
    muts = []
    m_gen = []
    m_ref = []
    m_alt = []
    for var in VCF(filepath):
        if var.is_snp or var.is_indel:
            row = [var.CHROM, var.POS, var.REF, var.ALT]
            ann = var.INFO.get("ANN").split(",")[0].split("|")
            row += ann
            m_gen.append(var.gt_types.tolist())
            m_ref.append(var.gt_ref_depths.tolist())
            m_alt.append(var.gt_alt_depths.tolist())
            muts.append(row)

    m_gen = np.array(m_gen)
    m_ref = np.array(m_ref)
    m_alt = np.array(m_alt)
    muts = pd.DataFrame(muts, columns=info)
    muts.index = muts.index.map(lambda x: f"mut{x}")
    cells = pd.DataFrame(index=cells)

    adata = ad.AnnData(np.zeros((len(muts), len(cells))))
    adata.obs = muts
    adata.var = cells
    adata.layers["genotype"] = m_gen
    adata.layers["total"] = m_ref + m_alt
    adata.layers["mutant"] = m_alt
    adata = adata.T
    return adata



import trisicell as tsc


def cardelino(adata, mode, kind, tree=None):
    """running Cardelino

    This method uses the prebuilt tree (either by bWES or scDNA) and
    then tries to cluster single-cells of scRNA based on the appearance
    of mutations detect in the tree.
    https://doi.org/10.1038/s41592-020-0766-3

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        Input data contains layers of mutant and total.
    mode : :obj:`str`
        Cardelino mode, either `fixed` or `free`
    kind : :obj:`str`
        Which dataset is this, either `brna` or `scrna`
    tree : :class:`pandas.DataFrame`
        Input prebuilt tree in the form of conflict-free matrix.

    Returns
    -------
    :class:`pandas.DataFrame`
        The assignment of single-cells to the clones of the tree.
    """

    try:
        import rpy2.robjects as robjects
        from rpy2.robjects.packages import importr
    except:
        raise SystemExit(
            "A problem was encountered importing `rpy2`. "
            "To run this `rpy2` and `R` need to be installed."
        )
    try:
        importr("cardelino")
    except:
        raise SystemExit(
            "A problem was encountered importing `cardelino` in R. "
            "To run this `cardelino` needs to be installed in R. "
            "Use the following lines to installed them.\n\n"
            "devtools::install_github('single-cell-genetics/cardelino',build_vignettes = FALSE)\n"
        )

    tsc.logg.info(f"running Cardelino")

    if kind == "brna":
        n_clone = 12
    elif kind == "scrna":
        n_clone = 25
    else:
        n_clone = 0

    A = adata.to_df(layer="mutant").T
    D = adata.to_df(layer="total").T
    if mode == "fixed":
        Config = tree.T

    robjects.pandas2ri.activate()
    A_r = robjects.conversion.py2rpy(A)
    D_r = robjects.conversion.py2rpy(D)
    if mode == "fixed":
        Config_r = robjects.conversion.py2rpy(Config)

    robjects.globalenv["A"] = A_r
    robjects.globalenv["D"] = D_r
    if mode == "fixed":
        robjects.globalenv["Config"] = Config_r

    cmd = f"""
    suppressPackageStartupMessages({{
        library(cardelino)
    }})
    A = as.matrix(A)
    D = as.matrix(D)
    """

    if mode == "free":
        cmd += f"""
        assignments <- clone_id(A, D, n_clone={n_clone})
        """
    elif mode == "fixed":
        cmd += f"""
        Config = as.matrix(Config)
        assignments <- clone_id(A, D, Config=Config)
        """
    else:
        raise ValueError("Wrong mode")

    cmd += f"""
    result <- assign_cells_to_clones(assignments$prob)
    result
    """

    # cmd = f"""
    # suppressPackageStartupMessages({{
    #     library(cardelino)
    # }})
    # vcf <- read_vcf(system.file('extdata', 'cell_example.mpileup.vcf.gz', package = 'cardelino'))
    # input_data <- get_snp_matrices(vcf)
    # canopy <- readRDS(system.file('extdata', 'canopy_results.example.rds', package = 'cardelino'))
    # Config <- canopy$tree$Z
    # rownames(Config) <- gsub('chr', '', gsub(':', '_', gsub('_.*', '', rownames(Config))))
    # assignments <- clone_id(input_data$A, input_data$D, Config) #mutant, total, tree
    # result <- assign_cells_to_clones(assignments$prob)
    # """

    result = robjects.r(cmd)

    return result



import numpy as np
import pytest
from _helpers import *

import trisicell as tsc


class TestScores:
    def test_ad(self):
        f1 = tsc.ul.get_file("trisicell.datasets/test/fp_0-fn_0-na_0.ground.CFMatrix")
        f2 = tsc.ul.get_file("trisicell.datasets/test/fp_1-fn_0.1-na_0.bnb.CFMatrix")
        grnd = tsc.io.read(f1)
        sol = tsc.io.read(f2)
        ad = tsc.tl.ad(grnd, sol)
        assert np.abs(ad - 0.9778) < 0.0001

    def test_dl(self):
        f1 = tsc.ul.get_file("trisicell.datasets/test/fp_0-fn_0-na_0.ground.CFMatrix")
        f2 = tsc.ul.get_file("trisicell.datasets/test/fp_1-fn_0.1-na_0.bnb.CFMatrix")
        grnd = tsc.io.read(f1)
        sol = tsc.io.read(f2)
        dl = tsc.tl.dl(grnd, sol)
        assert np.abs(dl - 0.9880) < 0.0001

    def test_mltd(self):
        f1 = tsc.ul.get_file("trisicell.datasets/test/fp_0-fn_0-na_0.ground.CFMatrix")
        f2 = tsc.ul.get_file("trisicell.datasets/test/fp_1-fn_0.1-na_0.bnb.CFMatrix")
        grnd = tsc.io.read(f1)
        sol = tsc.io.read(f2)
        mltd = tsc.tl.mltd(grnd, sol)
        assert np.abs(mltd[2] - 0.7800) < 0.0001

    def test_tpted(self):
        f1 = tsc.ul.get_file("trisicell.datasets/test/fp_0-fn_0-na_0.ground.CFMatrix")
        f2 = tsc.ul.get_file("trisicell.datasets/test/fp_1-fn_0.1-na_0.bnb.CFMatrix")
        grnd = tsc.io.read(f1)
        sol = tsc.io.read(f2)
        tpted = tsc.tl.tpted(grnd, sol)
        assert tpted == 0.99



import pytest
from _helpers import *

import trisicell as tsc


class TestSolvers:
    @skip_rpy2
    def test_simulate(self):
        df_in = tsc.datasets.simulate(
            n_cells=10, n_muts=10, n_clones=3, alpha=0.00001, beta=0.2, missing=0.1
        )
        is_cf = tsc.ul.is_conflict_free_gusfield(df_in)
        assert is_cf == False

    def test_scistree(self):
        df_in = tsc.datasets.test()
        df_out = tsc.tl.scistree(df_in, alpha=0.0000001, beta=0.1)
        is_cf = tsc.ul.is_conflict_free_gusfield(df_out)
        assert is_cf == True

    def test_scite(self):
        df_in = tsc.datasets.test()
        df_out = tsc.tl.scite(
            df_in, alpha=0.0000001, beta=0.1, n_restarts=3, n_iters=1000
        )
        is_cf = tsc.ul.is_conflict_free_gusfield(df_out)
        assert is_cf == True

    @skip_mpi4py
    def test_bnb(self):
        df_in = tsc.datasets.test()
        df_out = tsc.tl.bnb(df_in, bounding="simulated")
        is_cf = tsc.ul.is_conflict_free_gusfield(df_out)
        assert is_cf == True

    def test_phiscsb(self):
        df_in = tsc.datasets.test()
        df_out = tsc.tl.phiscsb(df_in, alpha=0.0000001, beta=0.1)
        is_cf = tsc.ul.is_conflict_free_gusfield(df_out)
        assert is_cf == True

    @skip_gurobi
    def test_phiscs_original(self):
        adata = tsc.datasets.acute_lymphocytic_leukemia2()
        adata.var["VAF"] = (
            2
            * adata.var["MutantCount"]
            / (adata.var["MutantCount"] + adata.var["ReferenceCount"])
        )
        df_out = tsc.tl.phiscs_original(
            adata.to_df(),
            alpha=0.001,
            beta=0.181749,
            delta=0.2,
            kmax=3,
            kel_weight=0,
            vaf_info=adata.var[["VAF"]],
        )
        is_cf = tsc.ul.is_conflict_free_gusfield(df_out)
        assert is_cf == True

    def test_booster(self):
        df_in = tsc.datasets.test()
        df_out = tsc.tl.booster(
            df_in,
            alpha=0.0000001,
            beta=0.1,
            solver="SCITE",
            sample_on="muts",
            sample_size=10,
            n_samples=20,
            begin_sample=0,
            n_jobs=1,
            time_out=10000,
            save_inter=False,
            dir_inter=".",
            base_inter=None,
            disable_tqdm=False,
            weight=5,
            no_subsampling=False,
            no_dependencies=False,
        )
        is_cf = tsc.ul.is_conflict_free_gusfield(df_out)
        assert is_cf == False



import pytest
from _helpers import *

import trisicell as tsc


class TestParitionFunction:
    def test_partition_function(self):
        df_in = tsc.datasets.test()
        probs = tsc.tl.partition_function(
            df_in,
            alpha=0.000001,
            beta=0.1,
            n_samples=100,
            n_batches=10,
            muts=["mut12"],
            cells=["cell6", "cell17"],
        )
        probs.mean(axis=1).round(4).values[0] >= 0.2
